# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

# In order to avoid circular dependencies, we use the following import order:
# - Containers
# - Contexts
# - Task
# - Materialize
# - Core:
#   * PipedagConfig (rather independent from other core)
#   * GroupNode (connected to task)
#   * VisualizationStyle (connected to GroupNode)
#   * Stage (needed by Flow)
#   * Result (needed by Flow)
#   * Flow (needs a lot of the above)
# - Engine
# - Backends

# isort: skip_file
from .container import (
    Blob,
    ExternalTableReference,
    RawSql,
    Schema,
    Table,
    materialize_table,
)
from .context import ConfigContext, StageLockContext
from .core.task import (
    Task,
    TaskGetItem,
)
from .materialize import (
    input_stage_versions,
    materialize,
)
from .materialize.materializing_task import AUTO_VERSION
from .core import (
    PipedagConfig,
    GroupNode,
    VisualizationStyle,
    Stage,
    Result,
    Flow,
)
from . import backend

__all__ = [
    "Flow",
    "Stage",
    "materialize",
    "input_stage_versions",
    "Table",
    "RawSql",
    "Blob",
    "ExternalTableReference",
    "Task",
    "TaskGetItem",
    "GroupNode",
    "VisualizationStyle",
    "Schema",
    "Result",
    "PipedagConfig",
    "ConfigContext",
    "StageLockContext",
    "AUTO_VERSION",
    "backend",
    "materialize_table",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.pipedag import ConfigContext


def compile_sql(query):
    engine = ConfigContext.get().store.table_store.engine
    return str(query.compile(engine, compile_kwargs={"literal_binds": True}))



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pathlib import Path

import pandas as pd
import sqlalchemy as sa

from pydiverse.pipedag import ConfigContext, Flow, Stage, Table, materialize
from tests.fixtures.instances import with_instances


@materialize(input_type=sa.Table, lazy=True)
def table_1(script_path: str):
    sql = Path(script_path).read_text(encoding="utf-8")
    return Table(sa.text(sql), name="table_1")


@materialize(input_type=sa.Table, lazy=True)
def table_2(script_path: str, dependent_table: Table):
    sql = Path(script_path).read_text(encoding="utf-8").replace("{{dependent}}", str(dependent_table.original))
    return Table(sa.text(sql), name="test_table2")


@materialize(input_type=pd.DataFrame, lazy=True)
def assert_result(df: pd.DataFrame):
    pd.testing.assert_frame_equal(df, pd.DataFrame({"coltab2": [24]}), check_dtype=False)


@with_instances("postgres", "mssql", "ibm_db2", per_user=True)
def test_sql_node():
    instance_name = ConfigContext.get().instance_name

    script_1_name = {
        "ibm_db2": "script1-db2.sql",
    }.get(instance_name, "script1.sql")
    script_2_name = "script2.sql"

    with Flow("FLOW") as flow:
        with Stage("schema1"):
            parent_dir = Path(__file__).parent
            tab1 = table_1(str(parent_dir / "sql_scripts" / script_1_name))
            tab2 = table_2(str(parent_dir / "sql_scripts" / script_2_name), tab1)
            assert_result(tab2)

    flow_result = flow.run()
    assert flow_result.successful


if __name__ == "__main__":
    test_sql_node()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import os
from pathlib import Path
from typing import Any

import pandas as pd
import pytest
import sqlalchemy as sa

from pydiverse.pipedag import Flow, Stage, Table, materialize
from pydiverse.pipedag.context import StageLockContext
from pydiverse.pipedag.core.config import PipedagConfig
from tests.fixtures.instances import with_instances

pytestmark = [with_instances("postgres")]


@pytest.fixture(
    scope="function",
    params=[
        "pipedag_complex.yaml",
        "pipedag_anchor.yaml",
    ],
)
def cfg_file_path(request):
    password_cfg_path = str(Path(__file__).parent / "postgres_password.yaml")

    old_environ = dict(os.environ)
    os.environ["POSTGRES_PASSWORD_CFG"] = password_cfg_path

    yield Path(__file__).parent / request.param

    os.environ.clear()
    os.environ.update(old_environ)


########################################################################################


dfA_source = pd.DataFrame(
    {
        "a": [0, 1, 2, 4],
        "b": [9, 8, 7, 6],
    }
)
dfA = dfA_source.copy()
input_hash = hash(str(dfA))


def has_new_input():
    global input_hash
    return input_hash


@materialize(nout=2, cache=has_new_input, version="1.0")
def input_task():
    global dfA
    return Table(dfA, "dfA"), Table(dfA, "dfB")


def has_copy_source_fresh_input(stage: Stage, attrs: dict[str, Any], pipedag_config: PipedagConfig):
    source = attrs["copy_source"]
    per_user = attrs["copy_per_user"]
    source_cfg = pipedag_config.get(instance=source, per_user=per_user)
    with source_cfg:
        _hash = source_cfg.store.table_store.get_stage_hash(stage)
    return _hash


@materialize(input_type=pd.DataFrame, cache=has_copy_source_fresh_input, version="1.0")
def copy_filtered_inputs(stage: Stage, attrs: dict[str, Any], pipedag_config: PipedagConfig):
    source = attrs["copy_source"]
    per_user = attrs["copy_per_user"]
    filter_cnt = attrs["copy_filter_cnt"]
    tbls = _get_source_tbls(source, per_user, stage, pipedag_config)
    ret = [Table(tbl.head(filter_cnt), name) for name, tbl in tbls.items()]
    return ret


def _get_source_tbls(source, per_user, stage, pipedag_config):
    source_cfg = pipedag_config.get(instance=source, per_user=per_user)
    with source_cfg:
        # This is just quick hack code to copy data from one pipeline instance to
        # another in a filtered way. It justifies actually a complete pydiverse
        # package called pydiverse.testdata. We want to achieve loose coupling by
        # pipedag transporting uninterpreted attrs with user code feeding the
        # attributes in testdata functionality
        engine = source_cfg.store.table_store.engine
        schema = source_cfg.store.table_store.get_schema(stage.name)
        meta = sa.MetaData()
        meta.reflect(bind=engine, schema=schema.name)
        tbls = {tbl.name: pd.read_sql_table(tbl.name, con=engine, schema=schema.name) for tbl in meta.tables.values()}
    return tbls


@materialize(input_type=pd.DataFrame, version="1.0")
def double_values(df: pd.DataFrame):
    return Table(df.transform(lambda x: x * 2))


@materialize(nout=2, input_type=sa.Table, lazy=True)
def extract_a_b(tbls: list[sa.Table]):
    a = [tbl for tbl in tbls if tbl.original.name == "dfa"][0]
    b = [tbl for tbl in tbls if tbl.original.name == "dfb"][0]
    return a, b


# noinspection PyTypeChecker
def get_flow(attrs: dict[str, Any], pipedag_config):
    with Flow("test_instance_selection") as flow:
        with Stage("stage_1") as stage:
            if not attrs["copy_filtered_input"]:
                a, b = input_task()
            else:
                tbls = copy_filtered_inputs(stage, attrs, pipedag_config)
                a, b = extract_a_b(tbls)
            a2 = double_values(a)

        with Stage("stage_2"):
            b2 = double_values(b)
            a3 = double_values(a2)

    return flow, b2, a3


def test_instance_selection(cfg_file_path):
    # At this point, an instance is chosen from multi-pipedag-instance
    # configuration file
    pipedag_config = PipedagConfig(cfg_file_path)
    cfg = pipedag_config.get(instance="full")

    flow, out1, out2 = get_flow(cfg.attrs, pipedag_config)

    with StageLockContext():
        result = flow.run(config=cfg)
        check_result(result, out1, out2)

    cfg = pipedag_config.get(instance="midi")

    flow, out1, out2 = get_flow(cfg.attrs, pipedag_config)

    with StageLockContext():
        result = flow.run(config=cfg)
        check_result(result, out1, out2, head=2)

    cfg = pipedag_config.get(instance="mini")

    flow, out1, out2 = get_flow(cfg.attrs, pipedag_config)

    with StageLockContext():
        result = flow.run(config=cfg)
        check_result(result, out1, out2, head=1)


def check_result(result, out1, out2, *, head=999):
    assert result.successful
    v_out1, v_out2 = result.get(out1), result.get(out2)
    pd.testing.assert_frame_equal(dfA_source.head(head) * 2, v_out1, check_dtype=False)
    pd.testing.assert_frame_equal(dfA_source.head(head) * 4, v_out2, check_dtype=False)


# In the future, the following test functions should be auto-generatable
# via pydiverse.pipetest library based on tags in pipedag_complex.yaml


@pytest.mark.slow5
def test_run_full_instance(cfg_file_path):
    pipedag_config = PipedagConfig(cfg_file_path)
    cfg = pipedag_config.get(instance="full")

    flow, out1, out2 = get_flow(cfg.attrs, pipedag_config)

    flow.run(config=cfg)


@pytest.mark.slow4
def test_run_midi_instance(cfg_file_path):
    pipedag_config = PipedagConfig(cfg_file_path)
    cfg = pipedag_config.get(instance="midi")

    flow, out1, out2 = get_flow(cfg.attrs, pipedag_config)

    flow.run(config=cfg)


@pytest.mark.slow3
def test_midi_instance_stages(cfg_file_path):
    pipedag_config = PipedagConfig(cfg_file_path)
    cfg = pipedag_config.get(instance="midi")

    flow, out1, out2 = get_flow(cfg.attrs, pipedag_config)

    flow.run(config=cfg, stages=["simple_flow_stage2"])


@pytest.mark.slow2
def test_run_mini_instance(cfg_file_path):
    pipedag_config = PipedagConfig(cfg_file_path)
    cfg = pipedag_config.get(instance="mini")

    flow, out1, out2 = get_flow(cfg.attrs, pipedag_config)

    flow.run(config=cfg)


@pytest.mark.slow1
def test_mini_instance_stages(cfg_file_path):
    pipedag_config = PipedagConfig(cfg_file_path)
    cfg = pipedag_config.get(instance="mini")

    flow, out1, out2 = get_flow(cfg.attrs, pipedag_config)

    flow.run(config=cfg, stages=["simple_flow_stage2"])



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import hashlib
import threading
import time
import warnings
from abc import ABC, abstractmethod
from typing import Any

import sqlalchemy as sa

from pydiverse.pipedag import ConfigContext, Stage
from pydiverse.pipedag.backend.lock.base import BaseLockManager, Lockable, LockState
from pydiverse.pipedag.backend.table.sql.ddl import (
    CreateSchema,
)
from pydiverse.pipedag.container import Schema
from pydiverse.pipedag.errors import LockError

try:
    from sqlalchemy import Connection, Engine
except ImportError:
    # For compatibility with sqlalchemy < 2.0
    from sqlalchemy.engine import Engine
    from sqlalchemy.engine.base import Connection


DISABLE_DIALECT_REGISTRATION = "__DISABLE_DIALECT_REGISTRATION"


class DatabaseLockManager(BaseLockManager):
    """Lock manager based on database locking mechanisms

    Many databases provide some kind of locking mechanism. Depending on the specific
    database technology, this allows us to implement locking on either the schema
    level (where each stage can be locked and unlocked individually), or only on
    the instance level (where the entire instance including all stages get locked and
    unlocked together).

    This lock manager uses the same database as the ``SQLTableStore``.
    No other configuration in the ``args`` section of the config file is required.

    We currently support the following databases:
    PostgreSQL, Microsoft SQL Server, IBM DB2.
    """

    __registered_dialects: dict[str, type["DatabaseLockManager"]] = {}
    _dialect_name: str

    def __new__(cls, engine: Engine, *args, **kwargs):
        if cls != DatabaseLockManager:
            return super().__new__(cls)

        # If calling DatabaseLockManager(engine), then we want to dynamically
        # instantiate the correct dialect specific subclass based on the engine dialect.
        dialect = engine.dialect.name
        dialect_specific_cls = DatabaseLockManager.__registered_dialects.get(dialect, cls)
        return super(DatabaseLockManager, dialect_specific_cls).__new__(dialect_specific_cls)

    @classmethod
    def _init_conf_(cls, config: dict[str, Any]):
        assert len(config) == 0, "DatabaseLockManager doesn't take any additional arguments"

        cfg_context = ConfigContext.get()
        return cls.init_from_config_context(cfg_context)

    @classmethod
    def init_from_config_context(cls, config_context: ConfigContext):
        from pydiverse.pipedag.backend.table import SQLTableStore

        table_store = config_context.store.table_store
        if not isinstance(table_store, SQLTableStore):
            raise TypeError("Table Store must be a subclass of SQLTableStore")

        engine = table_store.get_engine_for_locking()
        instance_id = config_context.instance_id
        lock_schema = table_store.get_lock_schema()
        create_lock_schema = not table_store.avoid_drop_create_schema

        return cls(engine, instance_id, lock_schema, create_lock_schema)

    def __init__(
        self,
        engine: Engine,
        instance_id: str,
        lock_schema: Schema | None = None,
        create_lock_schema: bool = True,
    ):
        super().__init__(logger_kwargs=dict(engine_url=engine.url, instance_id=instance_id, schema=lock_schema))

        self.engine = engine
        self.instance_id = instance_id
        self.lock_schema = lock_schema
        self.create_lock_schema = create_lock_schema

        # Stage level locking
        self.connection = None
        self.locks: dict[Lockable, Lock] = {}

        # Instance level locking
        self.il_lock: Lock | None = None
        self.il_locks: set[Lockable] = set()
        self.__il_threading_lock = threading.Lock()

        # Prepare database
        self.prepare()

    def __init_subclass__(cls, **kwargs):
        # Whenever a new subclass if DatabaseLockManager is defined, it must contain the
        # `_dialect_name` attribute. This allows us to dynamically instantiate it
        # when DatabaseLockManager SQLTableStore(engine, ...) based on the dialect
        # of the engine (see __new__).
        dialect_name = getattr(cls, "_dialect_name", None)
        if dialect_name is None:
            raise ValueError(
                "All subclasses of DatabaseLockManager must have a `_dialect_name`"
                f"attribute. But {cls.__name__}._dialect_name is None."
            )

        if dialect_name != DISABLE_DIALECT_REGISTRATION:
            if dialect_name in DatabaseLockManager.__registered_dialects:
                warnings.warn(f"Already registered a DatabaseLockManager for dialect {dialect_name}")
            DatabaseLockManager.__registered_dialects[dialect_name] = cls

    def prepare(self):
        pass

    def _prepare_create_lock_schema(self):
        # Create the lock schema
        if not self.create_lock_schema:
            return

        # If two lock managers do this concurrently, this action might fail. Thus,
        # we retry a second time to ensure the schema exists.
        with self.engine.connect() as conn:
            exception = None
            for _ in range(3):
                try:
                    with conn.begin():
                        conn.execute(CreateSchema(self.lock_schema, if_not_exists=True))
                    break
                except sa.exc.DBAPIError as e:
                    exception = e
                    time.sleep(0.1)
            else:
                raise exception

    def dispose(self):
        self.engine.dispose()
        super().dispose()

    @property
    def supports_stage_level_locking(self):
        raise NotImplementedError

    def acquire(self, lockable: Lockable):
        if self.supports_stage_level_locking:
            lock = self.get_lock_object(lockable)
            self.logger.info(f"Locking '{lockable}'")
            if not lock.acquire():
                raise LockError(f"Failed to acquire lock '{lockable}'")
            self.locks[lockable] = lock
            self.set_lock_state(lockable, LockState.LOCKED)
        else:
            with self.__il_threading_lock:
                if len(self.il_locks) == 0:
                    if self.il_lock is None:
                        self.il_lock = self.get_instance_level_lock()
                    self.logger.info("Locking at instance level")
                    if not self.il_lock.acquire():
                        raise LockError(f"Failed to acquire lock '{lockable}'")

                if lockable in self.il_locks:
                    raise LockError("Already acquired lock")
                self.il_locks.add(lockable)
                self.set_lock_state(lockable, LockState.LOCKED)

    def release(self, lockable: Lockable):
        if self.supports_stage_level_locking:
            if lockable not in self.locks:
                raise LockError(f"No lock '{lockable}' found.")

            self.logger.info(f"Unlocking '{lockable}'")
            self.locks[lockable].release()
            del self.locks[lockable]
            self.set_lock_state(lockable, LockState.UNLOCKED)
        else:
            with self.__il_threading_lock:
                if self.il_lock is None or not self.il_lock.locked:
                    raise LockError("Instance level lock must be acquired first")

                if lockable not in self.il_locks:
                    raise LockError(f"No lock '{lockable}' found.")
                self.il_locks.remove(lockable)
                self.set_lock_state(lockable, LockState.UNLOCKED)

                if len(self.il_locks) == 0:
                    self.logger.info("Unlocking at instance level")
                    self.il_lock.release()

    def get_lock_object(self, lockable: Lockable):
        raise NotImplementedError(f"Dialect '{self.engine.dialect.name}' not supported")

    def get_instance_level_lock(self):
        raise NotImplementedError(f"Dialect '{self.engine.dialect.name}' not supported")

    def lock_name(self, lock: Lockable):
        if isinstance(lock, Stage):
            return self.instance_id + "#" + lock.name
        elif isinstance(lock, str):
            return self.instance_id + "#" + lock
        else:
            raise NotImplementedError(f"Can't lock object of type '{type(lock).__name__}'")


class Lock(ABC):
    @abstractmethod
    def acquire(self) -> bool:
        """Acquire the lock"""

    @abstractmethod
    def release(self) -> bool:
        """Release the lock"""

    @property
    @abstractmethod
    def locked(self) -> bool:
        """Returns the lock status"""


class PostgresLock(Lock):
    """
    Locking based on advisory locks.
    https://www.postgresql.org/docs/current/functions-admin.html#FUNCTIONS-ADVISORY-LOCKS
    """

    def __init__(self, name: str, connection: Connection):
        self.name = name

        digest = hashlib.sha256(name.encode("utf-8")).digest()
        self._id = int.from_bytes(digest[:8], byteorder="big", signed=True)

        self._connection = connection
        self._locked = False

    def acquire(self) -> bool:
        result = self._connection.exec_driver_sql(f"SELECT pg_catalog.pg_advisory_lock({self._id})").scalar()
        result = False if result is False else True
        self._locked = result
        return result

    def release(self) -> bool:
        result = self._connection.exec_driver_sql(f"SELECT pg_catalog.pg_advisory_unlock({self._id})").scalar()
        self._locked = False
        return False if result is False else True

    @property
    def locked(self) -> bool:
        return self._locked


class PostgresLockManager(DatabaseLockManager):
    _dialect_name = "postgresql"
    supports_stage_level_locking = True

    def get_lock_object(self, lockable: Lockable):
        if self.connection is None:
            self.connection = self.engine.connect()

        name = self.lock_name(lockable)
        return PostgresLock(name, self.connection)

    def dispose(self):
        self.release_all()
        if self.connection is not None:
            self.connection.close()
        super().dispose()


class MSSqlLock(Lock):
    """
    Locking based on application resource locks.
    https://learn.microsoft.com/en-us/sql/relational-databases/system-stored-procedures/sp-getapplock-transact-sql?view=sql-server-ver15
    """

    def __init__(self, name: str, connection: Connection):
        self.name = name
        self._connection = connection
        self._locked = False

    def acquire(self) -> bool:
        result = self._connection.execute(
            sa.text(
                """
                DECLARE @result int
                EXEC @result =
                    sp_getapplock
                        @Resource = :lock_name,
                        @LockMode = 'Exclusive',
                        @LockOwner = 'Session'
                SELECT @result
                """
            ),
            {"lock_name": self.name},
        ).scalar()
        result = result >= 0
        self._locked = result
        return result

    def release(self) -> bool:
        result = self._connection.execute(
            sa.text(
                """
                DECLARE @result int
                EXEC @result =
                    sp_releaseapplock
                        @Resource = :lock_name,
                        @LockOwner = 'Session'
                SELECT @result
                """
            ),
            {"lock_name": self.name},
        ).scalar()
        result = result >= 0
        self._locked = False
        return result

    @property
    def locked(self) -> bool:
        return self._locked


class MSSqlLockManager(DatabaseLockManager):
    _dialect_name = "mssql"
    supports_stage_level_locking = True

    def get_lock_object(self, lockable: Lockable):
        if self.connection is None:
            self.connection = self.engine.connect()

        name = self.lock_name(lockable)
        return MSSqlLock(name, self.connection)


class DB2Lock(Lock):
    """
    Locking based on 'LOCK TABLE' statements.
    https://www.ibm.com/docs/en/db2-for-zos/13?topic=statements-lock-table

    These locks can only be released by committing a transaction. Therefor it is
    not suitable for fine grain locking.
    """

    def __init__(self, table: str, schema: str, engine: Engine):
        self._engine = engine
        self._connection = None
        self._locked = False

        self.table = table
        self.schema = schema

    def acquire(self) -> bool:
        table = self._engine.dialect.identifier_preparer.quote(self.table)
        schema = self._engine.dialect.identifier_preparer.format_schema(self.schema)

        # CREATE TABLE IF NOT EXISTS
        with self._engine.begin() as conn:
            conn.exec_driver_sql(
                f"""
                BEGIN
                    DECLARE CONTINUE HANDLER FOR SQLSTATE '42710' BEGIN END;
                    EXECUTE IMMEDIATE 'CREATE TABLE {schema}.{table} (x int)';
                END
                """
            )

        # LOCK TABLE
        if self._connection is None:
            self._connection = self._engine.connect()
            self._connection.begin()

        self._connection.exec_driver_sql(f"LOCK TABLE {schema}.{table} IN EXCLUSIVE MODE")

        self._locked = True
        return True

    def release(self) -> bool:
        if self._connection is None:
            return False

        self._connection.close()
        self._locked = False
        return True

    @property
    def locked(self) -> bool:
        return self._locked


class DB2LockManager(DatabaseLockManager):
    _dialect_name = "ibm_db_sa"
    supports_stage_level_locking = False

    def prepare(self):
        self._prepare_create_lock_schema()

    def get_instance_level_lock(self):
        return DB2Lock("instance_lock", self.lock_schema.get(), self.engine)



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause
import functools
import operator
import types
from dataclasses import dataclass

import polars as pl
import pytest
import sqlalchemy as sa
import structlog
from polars.testing import assert_frame_equal

from pydiverse.pipedag import Flow, Stage, materialize, materialize_table
from pydiverse.pipedag.context.context import CacheValidationMode, ConfigContext
from pydiverse.pipedag.errors import HookCheckException
from tests.fixtures.instances import DATABASE_INSTANCES, with_instances

try:
    import pydiverse.colspec as cs
    import pydiverse.common as pdc
    import pydiverse.transform as pdt
except ImportError:
    cs = types.ModuleType("pydiverse.colspec")
    fn = lambda *args, **kwargs: (lambda *args, **kwargs: None)  # noqa: E731
    cs.__getattr__ = lambda name: object if name in ["ColSpec", "Collection"] else fn
    pdt = types.ModuleType("pydiverse.colspec")
    pdt.Table = None
    pdt.verb = lambda fn: fn
    pdt.SqlAlchemy = None
    pdt.Polars = None
    pdc = None

try:
    import dataframely as dy
except ImportError:
    dy = None


pytestmark = [
    pytest.mark.pdtransform,
    with_instances(tuple(set(DATABASE_INSTANCES) - {"ibm_db2"})),
]

# ------------------------------------------------------------------------------------ #
#                                        SCHEMA                                        #
# ------------------------------------------------------------------------------------ #


class MyFirstColSpec(cs.ColSpec):
    a = cs.Integer(primary_key=True)
    b = cs.Int16()  # TODO: replace with cs.Int16 once ColSpec 0.2.2 is on conda-forge
    c = cs.String  # cs.Enum(["x", "y"], nullable=True)


class MySecondColSpec(cs.ColSpec):
    a = cs.Integer(primary_key=True)
    b = cs.Integer(min=1)
    c = cs.String(nullable=False)  # cs.Enum(["x", "y"], nullable=False)


@dataclass
class MyCollection(cs.Collection):
    first: MyFirstColSpec
    second: MySecondColSpec

    @cs.filter()
    def equal_primary_keys(self):
        return functools.reduce(
            operator.and_,
            (self.first[key] == self.second[key] for key in self.common_primary_keys()),
        )

    @cs.filter()
    def first_b_greater_second_b(self):
        return (self.first.b > self.second.b) | self.pk_is_null(self.first) | self.pk_is_null(self.second)


@dataclass
class SimpleCollection(cs.Collection):
    first: MyFirstColSpec
    second: MySecondColSpec


# ------------------------------------------------------------------------------------ #
#                                         TESTS                                        #
# ------------------------------------------------------------------------------------ #


def data_without_filter_without_rule_violation() -> tuple[pdt.Table, pdt.Table]:
    first = pdt.Table({"a": [1, 2, 3], "b": [1, 2, 3], "c": ["x", "y", None]}, name="first")
    second = pdt.Table({"a": [1, 2, 3], "b": [1, 2, 3], "c": ["x", "y", "x"]}, name="second")
    return first, second


def data_without_filter_with_rule_violation() -> tuple[pdt.Table, pdt.Table]:
    first = pdt.Table({"a": [1, 2, 1], "b": [1, 2, 3], "c": ["x", "y", None]}, name="first")
    second = pdt.Table({"a": [1, 2, 3], "b": [0, 1, 2], "c": [None, "y", "x"]}, name="second")
    return first, second


def data_with_filter_without_rule_violation() -> tuple[pdt.Table, pdt.Table]:
    first = pdt.Table({"a": [1, 2, 3], "b": [1, 1, 3], "c": ["x", "y", None]}, name="first")
    second = pdt.Table({"a": [2, 3, 4, 5], "b": [1, 2, 3, 4], "c": ["x", "y", "x", "x"]}, name="second")
    return first, second


def data_with_filter_with_rule_violation() -> tuple[pdt.Table, pdt.Table]:
    first = pdt.Table({"a": [1, 2, 3], "b": [1, 2, 3], "c": ["x", "y", None]}, name="first")
    second = pdt.Table(
        {"a": [2, 3, 4, 5, 6], "b": [0, 1, 2, 3, -1], "c": [None, "y", "y", "x", "z"]},
        name="second",
    )
    return first, second


@pytest.mark.skipif(cs.Collection is object, reason="ColSpec needs to be installed")
@pytest.mark.skipif(pdt.Table is None, reason="pydiverse.transform needs to be installed")
def test_dataclass():
    first, second = data_without_filter_without_rule_violation()
    c = SimpleCollection(first, second)
    assert_frame_equal(c.first >> pdt.export(pdt.Polars), first >> pdt.export(pdt.Polars))
    assert_frame_equal(c.second >> pdt.export(pdt.Polars), second >> pdt.export(pdt.Polars))


@materialize(nout=2)
def get_data(name: str):
    return globals()[f"data_{name}"]()


@materialize(nout=3, input_type=pdt.SqlAlchemy)
def exec_filter(c: cs.Collection):
    cfg = cs.config.Config.default

    def materialize_hook(tbl, table_prefix):
        return tbl >> materialize_table(table_prefix)

    store = ConfigContext.get().store.table_store
    cfg.dialect_name = store.engine.dialect.name
    cfg.materialize_hook = materialize_hook
    out, failure = c.filter(cfg=cfg, cast=True)
    return (
        out,
        SimpleCollection(
            **{name: f.invalid_rows >> pdt.alias(name + "_invalid_rows") for name, f in failure.__dict__.items()}
        ),
        SimpleCollection(**{name: f.counts() for name, f in failure.__dict__.items()}),
    )


# -------------------------------------- FILTER -------------------------------------- #


@pytest.mark.skipif(cs.Collection is object, reason="ColSpec needs to be installed")
@pytest.mark.skipif(pdt.Table is None, reason="pydiverse.transform needs to be installed")
def test_filter_without_filter_without_rule_violation():
    @materialize(input_type=pl.LazyFrame)
    def assertions(out, failure, failure_counts: dict[str, int]):
        first, second = data_without_filter_without_rule_violation()

        assert isinstance(out, SimpleCollection)
        assert_frame_equal(
            out.first.sort(by="a"),
            (first >> pdt.export(pdt.Polars(lazy=True))).cast(dict(b=pl.Int16)),
        )
        assert_frame_equal(out.second.sort(by="a"), second >> pdt.export(pdt.Polars(lazy=True)))
        assert failure.first.select(pl.len()).collect().item() == 0
        assert failure.second.select(pl.len()).collect().item() == 0

    with Flow() as flow:
        with Stage("s01"):
            c = SimpleCollection(*get_data("without_filter_without_rule_violation"))
        with Stage("s02"):
            out, failure, failure_counts = exec_filter(c)
            assertions(out, failure, failure_counts)

    flow.run()


@pytest.mark.skipif(cs.Collection is object, reason="ColSpec needs to be installed")
@pytest.mark.skipif(pdt.Table is None, reason="pydiverse.transform needs to be installed")
def test_filter_without_filter_with_rule_violation():
    @materialize(input_type=pl.LazyFrame)
    def assertions(out, failure, failure_counts: dict[str, int]):
        # first, second = data_without_filter_with_rule_violation()

        assert isinstance(out, SimpleCollection)
        assert len(out.first.collect()) == 1
        assert len(out.second.collect()) == 2
        assert failure_counts.first == {"_primary_key_": 2}
        assert failure_counts.second == {"b|min": 1, "c|nullability": 1}

    with Flow() as flow:
        with Stage("s01"):
            c = SimpleCollection(*get_data("without_filter_with_rule_violation"))
        with Stage("s02"):
            out, failure, failure_counts = exec_filter(c)
            assertions(out, failure, failure_counts)

    flow.run()


@pytest.mark.skipif(cs.Collection is object, reason="ColSpec needs to be installed")
@pytest.mark.skipif(pdt.Table is None, reason="pydiverse.transform needs to be installed")
def test_filter_with_filter_without_rule_violation():
    @materialize(input_type=pl.LazyFrame)
    def assertions(out, failure, failure_counts: dict[str, int]):
        # first, second = data_with_filter_without_rule_violation()

        assert isinstance(out, MyCollection)
        assert_frame_equal(
            out.first,
            pl.LazyFrame({"a": [3], "b": [3], "c": [None]}).cast(dict(b=pl.Int16, c=pl.String)),
        )
        assert_frame_equal(
            out.second,
            pl.LazyFrame({"a": [3], "b": [2], "c": ["y"]}).cast(dict(c=pl.String)),
        )
        assert failure_counts.first == {
            "equal_primary_keys": 1,
            "first_b_greater_second_b": 1,
        }
        assert failure_counts.second == {
            "equal_primary_keys": 2,
            "first_b_greater_second_b": 1,
        }

    with Flow() as flow:
        with Stage("s01"):
            c = MyCollection(*get_data("with_filter_without_rule_violation"))
        with Stage("s02"):
            out, failure, failure_counts = exec_filter(c)
            assertions(out, failure, failure_counts)

    flow.run()


@pytest.mark.skipif(cs.Collection is object, reason="ColSpec needs to be installed")
@pytest.mark.skipif(pdt.Table is None, reason="pydiverse.transform needs to be installed")
def test_filter_with_filter_with_rule_violation():
    @materialize(input_type=pl.LazyFrame)
    def assertions(out, failure, failure_counts: dict[str, int]):
        # first, second = data_with_filter_with_rule_violation()

        assert isinstance(out, MyCollection)
        assert_frame_equal(
            out.first,
            pl.LazyFrame({"a": [3], "b": [3], "c": [None]}).cast(dict(b=pl.Int16, c=pl.String)),
        )
        assert_frame_equal(
            out.second,
            pl.LazyFrame({"a": [3], "b": [1], "c": ["y"]}).cast(dict(c=pl.String)),
        )
        assert failure_counts.first == {"equal_primary_keys": 2}
        assert failure_counts.second == {
            "b|min": 2,
            "c|nullability": 1,
            # "c|dtype": 1,  # Enum currently doesn't work in pydiverse transform
            "equal_primary_keys": 2,
        }

    with Flow() as flow:
        with Stage("s01"):
            c = MyCollection(*get_data("with_filter_with_rule_violation"))
        with Stage("s02"):
            out, failure, failure_counts = exec_filter(c)
            assertions(out, failure, failure_counts)

    flow.run()


@pytest.mark.skipif(cs.Collection is object, reason="ColSpec needs to be installed")
@pytest.mark.skipif(
    dy is not None,
    reason="This test only works if dataframely is not installed since we test a fallback mechanism in polars hook",
)
@pytest.mark.parametrize(
    "with_filter, with_violation, validate_get_data",
    [(a, b, c) for a in [False, True] for b in [False, True] for c in [False, True]],
)
def test_annotations(with_filter: bool, with_violation: bool, validate_get_data: bool):
    if validate_get_data:

        @materialize(nout=2)
        def get_anno_data(
            name: str,
        ) -> tuple[MyFirstColSpec, MySecondColSpec]:
            return globals()[f"data_{name}"]()
    else:

        @materialize(nout=2)
        def get_anno_data(name: str) -> tuple[pdt.Table, pdt.Table]:
            return globals()[f"data_{name}"]()

    @materialize(input_type=pdt.Polars)
    def consumer(first: MyFirstColSpec, second: MySecondColSpec):
        assert [(c.name, c.dtype()) for c in first] == [
            ("a", pdc.Int64()),
            ("b", pdc.Int16() if validate_get_data or dy else pdc.Int64()),
            ("c", pdc.String()),
        ]
        assert [(c.name, c.dtype()) for c in second] == [
            ("a", pdc.Int64()),
            ("b", pdc.Int64()),
            ("c", pdc.String()),
        ]
        assert len(first >> pdt.export(pdt.Polars())) == 3
        assert len(second >> pdt.export(pdt.Polars())) in [3, 4, 5]

        if not validate_get_data and dy is None:
            # colspec will not do casting without dataframely
            # In case of validate_get_data, the type should already be
            # smallint in the database.
            first = first >> pdt.mutate(b=first.b.cast(pdt.Int16()))
        if not validate_get_data and with_violation:
            if with_filter:
                # this case does not occur for polars versions since they abort in cast
                MyFirstColSpec.validate(first)
            else:
                with pytest.raises(cs.exc.RuleValidationError, match="1 rules failed validation"):
                    MyFirstColSpec.validate(first)
            with pytest.raises(cs.exc.RuleValidationError, match="2 rules failed validation"):
                MySecondColSpec.validate(second)
        else:
            if not validate_get_data and dy is None:
                # colspec will not do casting without dataframely
                # In case of validate_get_data, the type should already be
                # smallint in the database.
                first = first >> pdt.mutate(b=first.b.cast(pdt.Int16()))
            assert MyFirstColSpec.is_valid(first)
            assert MySecondColSpec.is_valid(second)

    @materialize(input_type=pdt.Polars)
    def consumer2(first: MyFirstColSpec, second: MySecondColSpec):
        assert [(c.name, c.dtype()) for c in first] == [
            ("a", pdc.Int64()),
            ("b", pdc.Int16() if validate_get_data or dy else pdc.Int64()),
            ("c", pdc.String()),
        ]
        assert [(c.name, c.dtype()) for c in second] == [
            ("a", pdc.Int64()),
            ("b", pdc.Int64()),
            ("c", pdc.String()),
        ]
        assert len(first >> pdt.export(pdt.Polars())) == 3
        assert len(second >> pdt.export(pdt.Polars())) in [3, 4, 5]

    with Flow() as flow:
        name = f"with{'out' if not with_filter else ''}_filter_with{'out' if not with_violation else ''}_rule_violation"
        with Stage("s01"):
            first, second = get_anno_data(name)
            consumer(first, second)
        with Stage("s02"):
            consumer2(first, second)

    if with_violation and validate_get_data:
        # Validation at end of get_anno_data task fails
        with pytest.raises(
            HookCheckException,
            match="failed validation with MyFirstColSpec; Failure counts: {'b|min': 1, 'c|nullability': 1};"
            if with_filter
            else "{'_primary_key_': 2};",
        ):
            flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
    elif with_violation and not validate_get_data and with_filter:
        # # This only raises with polars operations since no cast is implemented without
        # # dataframely
        # with pytest.raises(
        #     RuntimeError,
        #     match="Failed to retrieve table '<Table 'get_anno_data",
        # ):
        ret = flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
        assert ret.successful
    else:
        ret = flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
        assert ret.successful


@pytest.mark.skipif(cs.Collection is object, reason="ColSpec needs to be installed")
@pytest.mark.skipif(
    dy is not None,
    reason="This test only works if dataframely is not installed since we test a fallback mechanism in polars hook",
)
@pytest.mark.parametrize(
    "with_filter, with_violation, validate_get_data",
    [(a, b, c) for a in [False, True] for b in [False, True] for c in [False, True]],
)
def test_annotations_not_fail_fast(with_filter: bool, with_violation: bool, validate_get_data: bool):
    if validate_get_data:

        @materialize(nout=2)
        def get_anno_data(
            name: str,
        ) -> tuple[MyFirstColSpec, MySecondColSpec]:
            return globals()[f"data_{name}"]()
    else:

        @materialize(nout=2)
        def get_anno_data(name: str) -> tuple[pdt.Table, pdt.Table]:
            return globals()[f"data_{name}"]()

    @materialize(input_type=pdt.Polars)
    def consumer(first: MyFirstColSpec, second: MySecondColSpec):
        assert [(c.name, c.dtype()) for c in first] == [
            ("a", pdc.Int64()),
            ("b", pdc.Int16() if validate_get_data or dy else pdc.Int64()),
            ("c", pdc.String()),
        ]
        assert [(c.name, c.dtype()) for c in second] == [
            ("a", pdc.Int64()),
            ("b", pdc.Int64()),
            ("c", pdc.String()),
        ]
        assert len(first >> pdt.export(pdt.Polars())) == 3
        assert len(second >> pdt.export(pdt.Polars())) in [3, 4, 5]

        if not validate_get_data and dy is None:
            # colspec will not do casting without dataframely
            # In case of validate_get_data, the type should already be
            # smallint in the database.
            first = first >> pdt.mutate(b=first.b.cast(pdt.Int16()))
        assert MyFirstColSpec.is_valid(first)
        assert MySecondColSpec.is_valid(second)

    @materialize(input_type=pdt.Polars)
    def consumer2(first: MyFirstColSpec, second: MySecondColSpec):
        assert [(c.name, c.dtype()) for c in first] == [
            ("a", pdc.Int64()),
            ("b", pdc.Int16() if validate_get_data or dy else pdc.Int64()),
            ("c", pdc.String()),
        ]
        assert [(c.name, c.dtype()) for c in second] == [
            ("a", pdc.Int64()),
            ("b", pdc.Int64()),
            ("c", pdc.String()),
        ]
        assert len(first >> pdt.export(pdt.Polars())) == 3
        assert len(second >> pdt.export(pdt.Polars())) in [3, 4, 5]

    with Flow() as flow:
        name = f"with{'out' if not with_filter else ''}_filter_with{'out' if not with_violation else ''}_rule_violation"
        with Stage("s01"):
            first, second = get_anno_data(name)
            consumer(first, second)
        with Stage("s02"):
            consumer2(first, second)

    with ConfigContext.get().evolve(fail_fast=False):
        result = flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
    if with_violation:
        assert not result.successful
    else:
        assert result.successful


@pytest.mark.skipif(cs.Collection is object, reason="ColSpec needs to be installed")
@pytest.mark.skipif(
    dy is not None,
    reason="This test only works if dataframely is not installed since we test a fallback mechanism in polars hook",
)
@pytest.mark.parametrize(
    "with_filter, with_violation, validate_get_data",
    [(a, b, c) for a in [False, True] for b in [False, True] for c in [False, True]],
)
def test_annotations_fault_tolerant(with_filter: bool, with_violation: bool, validate_get_data: bool):
    if validate_get_data:

        @materialize(nout=2)
        def get_anno_data(
            name: str,
        ) -> tuple[MyFirstColSpec, MySecondColSpec]:
            return globals()[f"data_{name}"]()
    else:

        @materialize(nout=2)
        def get_anno_data(name: str) -> tuple[pdt.Table, pdt.Table]:
            return globals()[f"data_{name}"]()

    @materialize(input_type=pdt.Polars)
    def consumer(first: MyFirstColSpec, second: MySecondColSpec):
        assert [(c.name, c.dtype()) for c in first] == [
            ("a", pdc.Int64()),
            (
                "b",
                pdc.Int16() if (validate_get_data and (not with_violation or with_filter)) or dy else pdc.Int64(),
            ),
            ("c", pdc.String()),
        ]
        assert [(c.name, c.dtype()) for c in second] == [
            ("a", pdc.Int64()),
            ("b", pdc.Int64()),
            ("c", pdc.String()),
        ]
        assert len(first >> pdt.export(pdt.Polars())) == 3
        assert len(second >> pdt.export(pdt.Polars())) in [3, 4, 5]

        if (not validate_get_data or (with_violation and not with_filter)) and dy is None:
            # colspec will not do casting without dataframely
            # In case of validate_get_data, the type should already be
            # smallint in the database.
            first = first >> pdt.mutate(b=first.b.cast(pdt.Int16()))
        if with_violation:
            if with_filter:
                MyFirstColSpec.validate(first)
                with pytest.raises(cs.exc.RuleValidationError, match="2 rules failed validation"):
                    MySecondColSpec.validate(second, cast=True)
            else:
                with pytest.raises(cs.exc.RuleValidationError, match="1 rules failed validation"):
                    MyFirstColSpec.validate(first)
                with pytest.raises(cs.exc.RuleValidationError, match="2 rules failed validation"):
                    MySecondColSpec.validate(second)
        else:
            assert MyFirstColSpec.is_valid(first)
            assert MySecondColSpec.is_valid(second)

    @materialize(input_type=pdt.Polars)
    def consumer2(first: MyFirstColSpec, second: MySecondColSpec):
        assert [(c.name, c.dtype()) for c in first] == [
            ("a", pdc.Int64()),
            (
                "b",
                pdc.Int16() if (validate_get_data and (not with_violation or with_filter)) or dy else pdc.Int64(),
            ),
            ("c", pdc.String()),
        ]
        assert [(c.name, c.dtype()) for c in second] == [
            ("a", pdc.Int64()),
            ("b", pdc.Int64()),
            ("c", pdc.String()),
        ]
        assert len(first >> pdt.export(pdt.Polars())) == 3
        assert len(second >> pdt.export(pdt.Polars())) in [3, 4, 5]

    with Flow() as flow:
        name = f"with{'out' if not with_filter else ''}_filter_with{'out' if not with_violation else ''}_rule_violation"
        with Stage("s01"):
            first, second = get_anno_data(name)
            consumer(first, second)
        with Stage("s02"):
            consumer2(first, second)

    with structlog.testing.capture_logs() as logs:
        with ConfigContext.get().evolve(table_hook_args=dict(polars=dict(fault_tolerant_annotation_action=True))):
            result = flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
    assert result.successful
    failures = [c for c in logs if c["event"] == "Failed to apply materialize annotation for table"]
    if with_violation and validate_get_data:
        assert len(failures) == 1 if with_filter else 2
        assert all("failed validation with My" in failure["exception"] for failure in failures)


@pytest.mark.skipif(cs.Collection is object, reason="ColSpec needs to be installed")
@pytest.mark.skipif(
    dy is not None,
    reason="This test only works if dataframely is not installed since we test a fallback mechanism in polars hook",
)
@pytest.mark.parametrize(
    "with_filter, with_violation, validate_get_data",
    [(a, b, c) for a in [False, True] for b in [False, True] for c in [False, True]],
)
def test_collections(with_filter: bool, with_violation: bool, validate_get_data: bool):
    CollectionType = MyCollection if with_filter else SimpleCollection

    if validate_get_data:

        @materialize()
        def get_anno_collection(name: str) -> CollectionType:
            first, second = globals()[f"data_{name}"]()
            return CollectionType(first=first, second=second)
    else:

        @materialize()
        def get_anno_collection(name: str):
            first, second = globals()[f"data_{name}"]()
            return CollectionType(first=first, second=second)

    @materialize(input_type=pdt.Polars)
    def consumer_collection(coll: CollectionType):
        # # collections are currently not passed on as annotations to individual tables
        # # thus no cast is happening
        assert [(c.name, c.dtype()) for c in coll.first] == [
            ("a", pdc.Int64()),
            ("b", pdc.Int64()),
            ("c", pdc.String()),
        ]
        assert [(c.name, c.dtype()) for c in coll.second] == [
            ("a", pdc.Int64()),
            ("b", pdc.Int64()),
            ("c", pdc.String()),
        ]
        assert len(coll.first >> pdt.export(pdt.Polars())) == 3
        assert len(coll.second >> pdt.export(pdt.Polars())) in [3, 4, 5]

        if with_violation:
            with pytest.raises(cs.exc.MemberValidationError, match="2 members failed validation"):
                coll.validate(cast=True)
        else:
            if with_filter:
                # it is not really without violation
                out, _ = coll.filter(cast=True)
                assert_frame_equal(
                    out.first >> pdt.export(pdt.Polars(lazy=True)),
                    pl.LazyFrame({"a": [3], "b": [3], "c": [None]}).cast(dict(b=pl.Int16, c=pl.String)),
                )
                assert_frame_equal(
                    out.second >> pdt.export(pdt.Polars(lazy=True)),
                    pl.LazyFrame({"a": [3], "b": [2], "c": ["y"]}),
                )
            else:
                assert coll.is_valid(cast=True)

    @materialize(input_type=pdt.Polars)
    def consumer2_collection(coll: CollectionType):
        # # collections are currently not passed on as annotations to individual tables
        # # thus no cast is happening
        assert [(c.name, c.dtype()) for c in coll.first] == [
            ("a", pdc.Int64()),
            ("b", pdc.Int64()),
            ("c", pdc.String()),
        ]
        assert [(c.name, c.dtype()) for c in coll.second] == [
            ("a", pdc.Int64()),
            ("b", pdc.Int64()),
            ("c", pdc.String()),
        ]
        assert len(coll.first >> pdt.export(pdt.Polars())) == 3
        assert len(coll.second >> pdt.export(pdt.Polars())) in [3, 4, 5]

    with Flow() as flow:
        name = f"with{'out' if not with_filter else ''}_filter_with{'out' if not with_violation else ''}_rule_violation"
        with Stage("s01"):
            collection = get_anno_collection(name)
            consumer_collection(collection)
        with Stage("s02"):
            consumer2_collection(collection)

    # # collections are currently not passed on as annotations to individual tables
    # # thus no cast is happening
    # if with_violation:
    #     from dataframely.exc import RuleValidationError
    #     with pytest.raises(RuleValidationError):
    #         flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
    # else:
    ret = flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
    assert ret.successful


@pytest.mark.skipif(cs.Collection is object, reason="ColSpec needs to be installed")
@pytest.mark.skipif(
    dy is not None,
    reason="This test only works if dataframely is not installed since we test a fallback mechanism in polars hook",
)
def test_type_mapping():
    @materialize(nout=2)
    def get_anno_data() -> tuple[MyFirstColSpec, MySecondColSpec]:
        return data_with_filter_without_rule_violation()

    @materialize(input_type=sa.Table)
    def consumer(first: MyFirstColSpec, second: MySecondColSpec):
        assert isinstance(first.c.b.type, sa.SmallInteger)
        assert isinstance(second.c.b.type, sa.BigInteger)
        assert not isinstance(second.c.b.type, sa.SmallInteger)

    with Flow() as flow:
        with Stage("s01"):
            first, second = get_anno_data()
            consumer(first, second)

    flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)


@pytest.mark.skipif(cs.Collection is object, reason="ColSpec needs to be installed")
@pytest.mark.parametrize(
    "with_filter, with_violation, validate_get_data",
    [(a, b, c) for a in [False, True] for b in [False, True] for c in [False, True]],
)
def test_annotations_sql(with_filter: bool, with_violation: bool, validate_get_data: bool):
    @materialize(nout=2)
    def load_raw_data(name: str):
        return globals()[f"data_{name}"]()

    if validate_get_data:

        @materialize(nout=2, input_type=pdt.SqlAlchemy)
        def get_anno_data_sql(first: MyFirstColSpec, second: MySecondColSpec) -> tuple[MyFirstColSpec, MySecondColSpec]:
            return first >> pdt.alias("first2"), second >> pdt.alias("second2")
    else:

        @materialize(nout=2, input_type=pdt.SqlAlchemy)
        def get_anno_data_sql(first: pdt.Table, second: pdt.Table) -> tuple[pdt.Table, pdt.Table]:
            return first >> pdt.alias("first2"), second >> pdt.alias("second2")

    def get_anno_data(name: str):
        first, second = load_raw_data(name)
        return get_anno_data_sql(first, second)

    @materialize(input_type=pdt.SqlAlchemy)
    def consumer(first: MyFirstColSpec, second: MySecondColSpec):
        assert [(c.name, c.dtype()) for c in first] == [
            ("a", pdc.Int64()),
            (
                "b",
                pdc.Int16() if validate_get_data and (not with_violation or with_filter) else pdc.Int64(),
            ),
            ("c", pdc.String()),
        ]
        assert [(c.name, c.dtype()) for c in second] == [
            ("a", pdc.Int64()),
            ("b", pdc.Int64()),
            ("c", pdc.String()),
        ]
        assert len(first >> pdt.export(pdt.Polars())) == 3
        assert len(second >> pdt.export(pdt.Polars())) in [3, 4, 5]

        # Conversion from sqlalchemy backend to polars backend always uses 64bit integer
        # intentionally
        first = first >> pdt.collect(pdt.Polars()) >> pdt.mutate(b=first.b.cast(pdt.Int16()))
        if not validate_get_data and with_violation:
            if with_filter:
                # this case does not occur for polars versions since they abort in cast
                MyFirstColSpec.validate(first)
            else:
                with pytest.raises(cs.exc.RuleValidationError, match="1 rules failed validation"):
                    MyFirstColSpec.validate(first)
            with pytest.raises(cs.exc.RuleValidationError, match="2 rules failed validation"):
                MySecondColSpec.validate(second)
        else:
            assert MyFirstColSpec.is_valid(first >> pdt.collect(pdt.Polars()))
            assert MySecondColSpec.is_valid(second >> pdt.collect(pdt.Polars()))

    @materialize(input_type=pdt.SqlAlchemy)
    def consumer2(first: MyFirstColSpec, second: MySecondColSpec):
        assert [(c.name, c.dtype()) for c in first] == [
            ("a", pdc.Int64()),
            (
                "b",
                pdc.Int16() if validate_get_data and (not with_violation or with_filter) else pdc.Int64(),
            ),
            ("c", pdc.String()),
        ]
        assert [(c.name, c.dtype()) for c in second] == [
            ("a", pdc.Int64()),
            ("b", pdc.Int64()),
            ("c", pdc.String()),
        ]
        assert len(first >> pdt.export(pdt.Polars())) == 3
        assert len(second >> pdt.export(pdt.Polars())) in [3, 4, 5]

    with Flow() as flow:
        name = f"with{'out' if not with_filter else ''}_filter_with{'out' if not with_violation else ''}_rule_violation"
        with Stage("s01"):
            first, second = get_anno_data(name)
            consumer(first, second)
        with Stage("s02"):
            consumer2(first, second)

    if with_violation and validate_get_data:
        # Validation at end of get_anno_data task fails
        with pytest.raises(
            HookCheckException,
            match="failed validation with MyFirstColSpec; Failure counts: {'b|min': 1, 'c|nullability': 1};"
            if with_filter
            else "{'_primary_key_': 2};",
        ):
            flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
    elif with_violation and not validate_get_data and with_filter:
        # # This only raises with polars operations since no cast is implemented without
        # # dataframely
        # with pytest.raises(
        #     RuntimeError,
        #     match="Failed to retrieve table '<Table 'get_anno_data",
        # ):
        ret = flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
        assert ret.successful
    else:
        ret = flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
        assert ret.successful



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pandas as pd
import pytest

from pydiverse.pipedag import AUTO_VERSION, Blob, Flow, Stage, Table
from pydiverse.pipedag.container import RawSql
from pydiverse.pipedag.materialize.core import materialize
from tests.fixtures.instances import with_instances
from tests.util import swallowing_raises

pytestmark = [with_instances("postgres"), with_instances("local_table_store")]


# Specific backends have tests in the test_table_hooks folder


def test_lazy_incompatible_with_auto_version():
    with pytest.raises(ValueError):

        @materialize(input_type=pd.DataFrame, version=AUTO_VERSION, lazy=True)
        def task(): ...


def test_missing_input_type_auto_version():
    with pytest.raises(ValueError):

        @materialize(version=AUTO_VERSION)
        def task(): ...


@with_instances("postgres")
def test_auto_version_illegal_return_types():
    @materialize(input_type=pd.DataFrame, version=AUTO_VERSION)
    def blob():
        return Blob(1), Table(pd.DataFrame())

    @materialize(input_type=pd.DataFrame, version=AUTO_VERSION)
    def raw_sql():
        return RawSql("..."), Table(pd.DataFrame())

    with Flow() as f:
        with Stage("auto_version"):
            _blob = blob()
            _raw_sql = raw_sql()

    with swallowing_raises(ValueError, match="Blob"):
        f.run(_blob)

    with swallowing_raises(ValueError, match="RawSql"):
        f.run(_raw_sql)


def test_auto_version_not_supported():
    import sqlalchemy as sa

    @materialize(input_type=sa.Table, version=AUTO_VERSION)
    def not_supported():
        return Table(pd.DataFrame({"x": [1, 2, 3, 4]}))

    with Flow() as f:
        with Stage("auto_version"):
            _ = not_supported()

    with swallowing_raises(TypeError, match="Auto versioning not supported"):
        f.run()


# TODO: Currently we only test that auto versioning actually works,
#       and that the task gets called the expected amount of times
#       in the polars hook tests.
#       Once we have support for auto versioning with pandas, we
#       might also want to put some tests into this file.



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from typing import TYPE_CHECKING, Callable, TypeVar, Union

if TYPE_CHECKING:
    from pydiverse.pipedag import Blob, Table
    from pydiverse.pipedag.materialize.store import BaseTableStore
    from pydiverse.pipedag.materialize.table_hook_base import TableHookResolver


def decorator_hint(decorator: Callable) -> Callable:
    # Used to fix incorrect type hints in pycharm
    return decorator


T = TypeVar("T")
CallableT = TypeVar("CallableT", bound=Callable)
StoreT = TypeVar("StoreT", bound="BaseTableStore")
TableHookResolverT = TypeVar("TableHookResolverT", bound="TableHookResolver")

# Materializable
MPrimitives = Union[int, float, bool, str]
MTypes = Union["Table", "Blob"]

BaseMaterializable = Union[MPrimitives, MTypes]
Materializable = Union[
    BaseMaterializable,
    dict[str, "Materializable"],
    list["Materializable"],
    tuple["Materializable", ...],
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import uuid
from collections.abc import Callable, Iterable
from contextlib import contextmanager
from functools import partial
from typing import TYPE_CHECKING, Any, overload

import sqlalchemy as sa

from pydiverse.common.util import deep_map
from pydiverse.common.util.hashing import stable_hash
from pydiverse.pipedag._typing import CallableT
from pydiverse.pipedag.container import Blob, RawSql, Table
from pydiverse.pipedag.context import ConfigContext, TaskContext
from pydiverse.pipedag.core.task import Task
from pydiverse.pipedag.materialize.materializing_task import (
    MaterializingTask,
    MaterializingTaskGetItem,
    UnboundMaterializingTask,
)
from pydiverse.pipedag.util.json import PipedagJSONEncoder

if TYPE_CHECKING:
    from pydiverse.pipedag import Stage


@overload
def materialize(
    *,
    name: str | None = None,
    input_type: type | None = None,
    version: str | None = None,
    cache: Callable[..., Any] | None = None,
    lazy: bool = False,
    nout: int = 1,
    add_input_source: bool = False,
    ordering_barrier: bool | dict[str, Any] = False,
    call_context: Callable[[], Any] | None = None,
) -> Callable[[CallableT], CallableT | UnboundMaterializingTask]: ...


@overload
def materialize(fn: CallableT, /) -> CallableT | UnboundMaterializingTask: ...


def materialize(
    fn: CallableT | None = None,
    *,
    name: str | None = None,
    input_type: type | None = None,
    version: str | None = None,
    cache: Callable[..., Any] | None = None,
    lazy: bool = False,
    group_node_tag: str | None = None,
    nout: int = 1,
    add_input_source: bool = False,
    ordering_barrier: bool | dict[str, Any] = False,
    call_context: Callable[[], Any] | None = None,
):
    """Decorator to create a task whose outputs get materialized.

    This decorator takes a class and turns it into a :py:class:`MaterializingTask`.
    This means, that this function can only be used as part of a flow.
    Any outputs it produces get written to their appropriate storage backends
    (either the table or blob store). Additionally, any :py:class:`Table`
    or :py:class:`Blob` objects this task receives as an input get replaced
    with the appropriate object retrieved from the storage backend.
    In other words: All outputs from a materializing task get written to the store,
    and all inputs get retrieved from the store.

    Because of how caching is implemented, task inputs and outputs must all
    be "materializable". This means that they can only contain objects of
    the following types:
    ``dict``, ``list``, ``tuple``,
    ``int``, ``float``, ``str``, ``bool``, ``None``,
    ``datetime.date``, ``datetime.datetime``, ``pathlib.Path``,
    :py:class:`Table`, :py:class:`RawSql`, :py:class:`Blob`, or :py:class:`Stage`.

    :param fn:
        The function that gets executed by this task.
    :param name:
        The name of this task.
        If no `name` is provided, the name of `fn` is used instead.
    :param input_type:
        The data type as which to retrieve table objects from the store.
        All tables passed to this task get loaded from the table store and converted
        to this type. See :doc:`Table Backends </table_backends>` for
        more information.
    :param version:
        The version of this task.
        Unless the task is lazy, you always need to manually change this
        version number when you change the implementation to ensure that the
        task gets executed and the cache flushed.

        If the `version` is ``None`` and the task isn't lazy, then the task always
        gets executed, and all downstream tasks get invalidated.
    :param cache:
        An explicit cache function used to determine cache validity of the task inputs.

        This function gets called every time before the task gets executed.
        It gets called with the same arguments as the task.

        An explicit function for validating cache validity. If the output
        of this function changes while the source parameters are the same (e.g.
        the source is a filepath and `cache` loads data from this file), then the
        cache will be deemed invalid and is not used.
    :param lazy:
        Whether this task is lazy or not.

        Unlike a normal task, lazy tasks always get executed. However, if a lazy
        task produces a lazy table (e.g. a SQL query), the table store checks if
        the same query has been executed before. If this is the case, then the
        query doesn't get executed, and instead, the table gets copied from the cache.

        This behaviour is very useful, because you don't need to manually bump
        the `version` of a lazy task. This only works because for lazy tables
        generating the query is very cheap compared to executing it.
    :param group_node_tag:
        Set a tag that may add this task to a configuration based group node.
    :param nout:
        The number of objects returned by the task.
        If set, this allows unpacking and iterating over the results from the task.
    :param add_input_source:
        If true, Table and Blob objects are provided as tuple together with the
        dematerialized object. ``task(a=tbl)`` will result in
        ``a=(dematerialized_tbl, tbl)`` in the task.
    :param ordering_barrier:
        If true, the task will be surrounded by a GroupNode(ordering_barrier=True).
        If a dictionary is provided, it is surrounded by a
        GroupNode(**ordering_barrier). This allows passing style, style_tag, and label
        arguments.
    :param call_context:
        An optional context manager function that is opened before the task or its
        optional cache function is called and closed afterward.

    Example
    -------

    ::

        @materialize(version="1.0", input_type=pd.DataFrame)
        def task(df: pd.DataFrame) -> pd.DataFrame:
            df = ...  # Do something with the dataframe
            return Table(df)

        @materialize(lazy=True, input_type=sa.Table)
        def lazy_task(tbl: sa.Alias) -> sa.Table:
            query = sa.select(tbl).where(...)
            return Table(query)

    You can use the `cache` argument to specify an explicit cache function.
    In this example we define a task that reads a dataframe from a csv file.
    Without specifying a cache function, this task would only get rerun when the
    `path` argument changes. Instead, we define a function that calculates a
    hash based on the contents of the csv file, and pass it to the `cache` argument
    of ``@materialize``. This means that the task now gets invalidated and rerun if
    either the `path` argument changes or if the content of the file pointed
    to by `path` changes::

        import hashlib

        def file_digest(path):
            with open(path, "rb") as f:
                return hashlib.file_digest(f, "sha256").hexdigest()

        @materialize(version="1.0", cache=file_digest)
        def task(path):
            df = pd.read_scv(path)
            return Table(df)

    Setting nout allows you to use unpacking assignment with the task result::

        @materialize(version="1.0", nout=2)
        def task():
            return 1, 2

        # Later, while defining a flow, you can use unpacking assignment
        # because you specified that the task returns 2 objects (nout=2).
        one, two = task()

    If a task returns a dictionary or a list, you can use square brackets to
    explicitly wire individual values / elements to task inputs::

        @materialize(version="1.0")
        def task():
            return {
                "x": [0, 1],
                "y": [2, 3],
            }

        # Later, while defining a flow
        # This would result in another_task being called with ([0, 1], 3).
        task_out = task()
        another_task(task_out["x"], task_out["y"][1])
    """
    if fn is None:
        return partial(
            materialize,
            name=name,
            input_type=input_type,
            version=version,
            cache=cache,
            lazy=lazy,
            group_node_tag=group_node_tag,
            nout=nout,
            add_input_source=add_input_source,
            ordering_barrier=ordering_barrier,
            call_context=call_context,
        )

    return UnboundMaterializingTask(
        fn,
        name=name,
        input_type=input_type,
        version=version,
        cache=cache,
        lazy=lazy,
        group_node_tag=group_node_tag,
        nout=nout,
        add_input_source=add_input_source,
        ordering_barrier=ordering_barrier,
        call_context=call_context,
    )


@overload
def input_stage_versions(
    *,
    name: str | None = None,
    input_type: type | None = None,
    version: str | None = None,
    cache: Callable[..., Any] | None = None,
    lazy: bool = False,
    group_node_tag: str | None = None,
    nout: int = 1,
    add_input_source: bool = False,
    ordering_barrier: bool | dict[str, Any] = True,
    call_context: Callable[[], Any] | None = None,
    include_views=True,
    lock_source_stages=True,
    pass_args: Iterable[str] = tuple(),
) -> Callable[[CallableT], CallableT | UnboundMaterializingTask]: ...


@overload
def input_stage_versions(fn: CallableT, /) -> CallableT | UnboundMaterializingTask: ...


def input_stage_versions(
    fn: CallableT | None = None,
    *,
    name: str | None = None,
    input_type: type | None = None,
    version: str | None = None,
    cache: Callable[..., Any] | None = None,
    lazy: bool = False,
    group_node_tag: str | None = None,
    nout: int = 1,
    add_input_source: bool = False,
    ordering_barrier: bool | dict[str, Any] = True,
    call_context: Callable[[], Any] | None = None,
    include_views=True,
    lock_source_stages=True,
    pass_args: Iterable[str] = tuple(),
):
    """
    A decorator that marks a function as a task that receives tables from two versions
    of the same stage. This can either be the currently active schema of this stage
    before schema swapping, or it can be an active schema of another pipeline instance.

    The arguments to the resulting task are significantly different from a task created
    with `@materialize` decorator. An arbitrary expression with lists, tuples, dicts,
    and Table/Blob/RawSQL references can be provided. This decorator will only collect
    tables referenced and will pass them as one dictionary per version to the task. It
    maps table names to some reference for the respective stage version in the form
    specified by input_type parameter.

    When a ConfigContext object is passed as toplevel parameter or named parameter in
    the wiring call of decorated task, it is used for referencing the second stage
    version. The idea is that the other configuration is just another pipeline instance
    accessible with the same SQLAlchemy engine of the table store. It may work also if
    this is not the case, but it may be harder for the task to process the dictionaries
    it receives. Currently, it is not allowed to pass more than one ConfigContext
    object.

    When no parameters (except for an optional ConfigContext) is passed, all currently
    existing tables of both stage versions (or their respective schema) are passed in
    the dictionaries to the task.

    With pass_blobs=True, two additional dictionaries are passed to the task at the end.
    In addition to the arguments mentioned above, pass_args can be used to feed defined
    keyword arguments through to the decorated task.

    :param fn:
        The function that gets executed by this task.
    :param name:
        The name of this task.
        If no `name` is provided, the name of `fn` is used instead.
    :param input_type:
        The data type as which to retrieve table objects from the store.
        All tables passed to this task get loaded from the table store and converted
        to this type. See :doc:`Table Backends </table_backends>` for
        more information.
    :param version:
        The version of this task.
        Unless the task is lazy, you always need to manually change this
        version number when you change the implementation to ensure that the
        task gets executed and the cache flushed.

        If the `version` is ``None`` and the task isn't lazy, then the task always
        gets executed, and all downstream tasks get invalidated.
    :param cache:
        An explicit cache function used to determine cache validity of the task inputs.

        This function gets called every time before the task gets executed.
        It gets called with the same arguments as the task.

        An explicit function for validating cache validity. If the output
        of this function changes while the source parameters are the same (e.g.
        the source is a filepath and `cache` loads data from this file), then the
        cache will be deemed invalid and is not used.
    :param lazy:
        Whether this task is lazy or not.

        Unlike a normal task, lazy tasks always get executed. However, if a lazy
        task produces a lazy table (e.g. a SQL query), the table store checks if
        the same query has been executed before. If this is the case, then the
        query doesn't get executed, and instead, the table gets copied from the cache.

        This behaviour is very useful, because you don't need to manually bump
        the `version` of a lazy task. This only works because for lazy tables
        generating the query is very cheap compared to executing it.
    :param group_node_tag:
        Set a tag that may add this task to a configuration based group node.
    :param nout:
        The number of objects returned by the task.
        If set, this allows unpacking and iterating over the results from the task.
    :param add_input_source:
        If true, Table and Blob objects are provided as tuple together with the
        dematerialized object. ``task(a=tbl)`` will result in
        ``a=(dematerialized_tbl, tbl)`` in the task.
    :param ordering_barrier:
        If true, the task will be surrounded by a GroupNode(ordering_barrier=True).
        If a dictionary is provided, it is surrounded by a
        GroupNode(**ordering_barrier). This allows passing style, style_tag, and label
        arguments.
        Attention: In contrast to @materialize, the default value is True.
    :param call_context:
        An optional context manager function that is opened before the task or its
        optional cache function is called and closed afterward.
    :param include_views:
        In case no explicit table references are given, if true, include views when
        collecting all tables of both stage versions.
    :param lock_source_stages:
        If true, lock the other stage version when a ConfigContext object is passed to
        this task.
    :param pass_args
        A list of named arguments that would be passed from the call to the task
        function. By default, no arguments are passed, and just tables are extracted.

    Example
    -------

    ::

        @input_stage_versions(lazy=True, input_type=sa.Table)
        def validate_stage(
            transaction: dict[str, sa.Alias],
            other: dict[str, sa.Alias],
        ):
            compare_tables(transaction, other)

        def get_flow():
            with Flow() as f:
                a = produce_a_table()
                validate_stage()  # implicitly receives all tables written before
    """
    from pydiverse.pipedag import Stage

    forward_args = dict(
        name=name,
        input_type=input_type,
        version=version,
        cache=cache,
        lazy=lazy,
        group_node_tag=group_node_tag,
        nout=nout,
        add_input_source=add_input_source,
        ordering_barrier=ordering_barrier,
        call_context=call_context,
        include_views=include_views,
        lock_source_stages=lock_source_stages,
        pass_args=pass_args,
    )

    if fn is None:
        return partial(input_stage_versions, **forward_args)

    class OtherStageLockManager:
        def __init__(self):
            self.stages = None
            self.lock_manager = None

        @contextmanager
        def __call__(self):
            yield
            self.release_all()

        def release_all(self):
            # clear locks after either cache_fn or cache_fn and task function were
            # called
            if self.lock_manager is not None:
                for stage in sorted(self.stages, reverse=True):
                    self.lock_manager.release(stage.name)
                self.lock_manager.dispose()
                self.lock_manager = None

        def memorize(self, stages, lock_manager):
            self.release_all()
            self.stages = stages
            self.lock_manager = lock_manager

    other_stage_lock_manager = OtherStageLockManager()

    def cache_fn(*args, **kwargs):
        user_cache_fn = cache
        if user_cache_fn is None:
            # we don't care about the cache function, but we still need to cache the
            # arguments that are also used to call the task itself
            def user_cache_fn(*args, **kwargs):
                return None

        ret, cfg2, stages, lock_manager = _call_fn(
            user_cache_fn,
            args,
            kwargs,
            input_type=None,
            is_dematerialized=False,
            lock_source_stages=lock_source_stages,
        )
        other_stage_lock_manager.memorize(stages, lock_manager)
        if cfg2 is None:
            return uuid.uuid4().hex  # return random hash to ensure invalidation
        # get a hash of all task outputs of all relevant stage from other pipeline
        # instance
        ret = stable_hash(ret)
        for stage in stages:
            try:
                ret = ret + cfg2.store.table_store.get_stage_hash(stage)
            except sa.exc.ProgrammingError:
                # this happens for example if the other instance did not run, yet
                return uuid.uuid4().hex  # return random hash to ensure invalidation
        return ret

    # extract @materialize arguments from forward_args via __code__.co_varnames:
    n_args = materialize.__code__.co_argcount + materialize.__code__.co_kwonlyargcount
    arg_names = materialize.__code__.co_varnames[0:n_args]
    materialize_args = {k: v for k, v in forward_args.items() if k in arg_names}
    materialize_args["add_input_source"] = True  # we need the names
    materialize_args["cache"] = cache_fn
    if materialize_args.get("name") is None:
        materialize_args["name"] = getattr(fn, "__name__", "input_stage_versions-task")
    materialize_args["call_context"] = other_stage_lock_manager

    # For every input_state_versions-Task, we create a materialize-Task and give it the
    # same name as the input_state_versions-Task should get. The @materialize decorator
    # will dematerialize any Table/Blob/RawSQL objects passed to the
    # input_state_versions-Task. We pass add_input_source=True, so we still receive the
    # Table/Blob/RawSQL reference objects, so we can use them to dematerialize the
    # second stage version. Both are arranged in dictionaries indexable by table name or
    # blob name and handed over to the function annotated with @input_stage_versions.
    @materialize(**materialize_args)
    def task(*args, **kwargs):
        # the locks are already acquired by cache_fn if it is called
        lock_stages = False if other_stage_lock_manager.lock_manager is not None else lock_source_stages
        ret, _, _, _ = _call_fn(
            fn,
            args,
            kwargs,
            input_type=input_type,
            is_dematerialized=True,
            lock_source_stages=lock_stages,
        )
        return ret

    def _call_fn(
        _fn: Callable[..., Any],
        args: tuple,
        kwargs: dict[str, Any],
        *,
        input_type,
        is_dematerialized,
        lock_source_stages,
    ):
        _task = TaskContext.get().task
        stage = _task.stage
        pass_kwargs = {k: v for k, v in kwargs.items() if k in set(pass_args)}

        (
            stages,
            table_dict,
            blob_dict,
            raw_sqls,
            other_cfg,
            any_data_arguments_collected,
        ) = _collect_arguments(args, kwargs, is_dematerialized)
        stages.add(stage)

        # The connection to the other stage version configuration can either be the same
        # (in this case, we provide the active stage cache before swapping) or another
        # pipeline instance (in this case, we also take the active stage cache schema
        # of the same stage in the other instance)
        cfg1 = ConfigContext.get()
        cfg2 = other_cfg
        if cfg2 is None:
            cfg2 = cfg1
        stage2 = Stage(
            stage.name,
            stage.materialization_details,
            stage.group_node_tag,
            force_committed=True,
        )
        stage2.id = stage.id
        if lock_source_stages and other_cfg is not None:
            lock_manager = cfg2.create_lock_manager()
        else:
            lock_manager = None
            lock_source_stages = False

        transaction_table_dict, other_table_dict = {}, {}
        transaction_blob_dict, other_blob_dict = {}, {}
        try:
            # lock the source stages if they are read from another instance
            if lock_source_stages:
                for stage in sorted(stages):
                    lock_manager.acquire(stage.name)
            # cnt[0] counts the number of arguments in args and kwargs. pass_args within
            # kwargs are neglected.
            # cnt[0] is at least 2 due to deep_map calls on args and kwargs. If a
            # ConfigContext object is passed as argument, cfg_cnt=1 and cnt=3.
            if not any_data_arguments_collected:
                # dematerialize all tables and optionally views of both stage versions

                def _dematerialize_all(cfg, stage):
                    table_dict = {}
                    if include_views:
                        # also include table aliases
                        tbls1 = cfg.store.table_store.get_objects_in_stage(stage)
                    else:
                        # won't detect aliases
                        tbls1 = cfg.store.table_store.get_table_objects_in_stage(stage, include_views=include_views)
                    for name in tbls1:
                        tbl = Table(name=name)
                        tbl.stage = stage
                        if include_views:
                            (
                                tbl.name,
                                tbl.external_schema,
                            ) = cfg.store.table_store.resolve_alias(tbl, stage.current_name)
                            if not cfg.store.table_store.has_table_or_view(tbl.name, tbl.external_schema):
                                continue  # skip because it is neither view nor alias
                        table_dict[tbl.name] = cfg.store.dematerialize_item(tbl, input_type)
                    return table_dict

                transaction_table_dict = _dematerialize_all(cfg1, stage)
                other_table_dict = _dematerialize_all(cfg2, stage2)
            else:

                def _dematerialize_versions(
                    ref: Table | Blob,
                    obj,
                    transaction_dict: dict[str, Any],
                    other_dict: dict[str, Any],
                    cfg2: ConfigContext,
                    stage: Stage,
                ):
                    key = ref.name if ref.stage == stage else f"{ref.stage.name}.{ref.name}"
                    transaction_dict[key] = obj
                    # load committed version of stage (either same pipeline instance or
                    # other)
                    tbl2 = ref.copy_without_obj()
                    tbl2.stage = Stage(
                        ref.stage.name,
                        ref.stage.materialization_details,
                        ref.stage.group_node_tag,
                        force_committed=True,
                    )
                    tbl2.stage.id = ref.stage.id
                    try:
                        store2 = cfg2.store.table_store
                        # Only try to dematerialize if an object with the correct name
                        # exists. Otherwise, we might run into retry loops.
                        if not isinstance(tbl2, Table) or tbl2.name.lower() in {
                            s.lower() for s in store2.get_objects_in_stage(tbl2.stage)
                        }:
                            other_dict[key] = cfg2.store.dematerialize_item(tbl2, input_type)
                        else:
                            other_dict[key] = (
                                f"Table not found in other stage version: {PipedagJSONEncoder().encode(tbl2)}"
                            )
                    except Exception as e:
                        _task.logger.error(
                            "Failed to dematerialize table from other stage version",
                            table=ref,
                            stage=tbl2.stage,
                            instance=cfg2.instance_name,
                            instance_id=cfg2.instance_id,
                            cause=str(e),
                        )
                        other_dict[key] = e

                for tbl, obj in table_dict.items():
                    _dematerialize_versions(tbl, obj, transaction_table_dict, other_table_dict, cfg2, stage)
                for blob, obj in blob_dict.items():
                    _dematerialize_versions(blob, obj, transaction_blob_dict, other_blob_dict, cfg2, stage)
                for raw_sql in raw_sqls:
                    for tbl, obj in raw_sql.loaded_tables:
                        _dematerialize_versions(
                            tbl,
                            obj,
                            transaction_table_dict,
                            other_table_dict,
                            cfg2,
                            stage,
                        )

            optional_cfg = [] if other_cfg is None else [other_cfg]

            # restructure arbitrary arguments referencing Tables/Blobs/RawSQL into 2 or
            # 4 dictionaries providing references only to tables and optionally blobs
            if len(transaction_blob_dict) > 0:
                ret = _fn(
                    transaction_table_dict,
                    other_table_dict,
                    transaction_blob_dict,
                    other_blob_dict,
                    *optional_cfg,
                    **pass_kwargs,
                )
            else:
                ret = _fn(
                    transaction_table_dict,
                    other_table_dict,
                    *optional_cfg,
                    **pass_kwargs,
                )
        except Exception as e:
            if lock_source_stages:
                for stage in sorted(stages, reverse=True):
                    lock_manager.release(stage.name)
            raise e
        return ret, cfg2, stages, lock_manager

    def _collect_arguments(args, kwargs, is_dematerialized):
        remaining_kwargs = {k: v for k, v in kwargs.items() if k not in set(pass_args)}
        table_dict, blob_dict, raw_sqls = {}, {}, []
        stages = set()
        other_cfg = [None]  # list to allow modification in visitor
        cnt = [0]
        cfg_cnt = [0]
        str_dict_keys = [0]

        def visitor(x):
            if isinstance(x, dict):
                str_dict_keys[0] += sum(isinstance(k, str) for k in x.keys())
            if not isinstance(x, (list, tuple, dict)):
                cnt[0] += 1  # count non-collection arguments
            if isinstance(x, ConfigContext):
                if cfg_cnt[0] > 0:
                    raise ValueError(
                        "Only one ConfigContext object is allowed when calling a task "
                        "decorated with @input_stage_versions."
                    )
                other_cfg[0] = x
                cfg_cnt[0] += 1
            if is_dematerialized:
                if isinstance(x, tuple) and len(x) == 2:
                    # with the add_input_source parameter of @materialize, we receive
                    # tuples of the reference object (Table/Blob/RawSQL) together with
                    # the dematerialized object
                    obj, ref = x
                else:
                    return x
            else:
                obj, ref = x, x
            if isinstance(ref, (Table, Blob, RawSql)):
                stages.add(ref.stage)
            if isinstance(ref, Table):
                table_dict[ref] = obj
            elif isinstance(ref, Blob):
                blob_dict[ref] = obj
            elif isinstance(ref, RawSql):
                # ref and obj are identical for RawSql
                raw_sqls.append(ref)
            return x

        deep_map(args, visitor)
        deep_map(remaining_kwargs, visitor)
        # dictionary keys don't count since we like to be f(config=cfg) to be considered
        # as not any_data_arguments_collected
        any_data_arguments_collected = cnt[0] > cfg_cnt[0] + str_dict_keys[0]
        return (
            stages,
            table_dict,
            blob_dict,
            raw_sqls,
            other_cfg[0],
            any_data_arguments_collected,
        )

    return task


def _get_output_from_store(
    task: MaterializingTask | MaterializingTaskGetItem,
    as_type: type,
    ignore_position_hashes: bool = False,
) -> Any:
    """Helper to retrieve task output from store"""
    from pydiverse.pipedag.context.run_context import DematerializeRunContext
    from pydiverse.pipedag.materialize.store import dematerialize_output_from_store

    root_task = task if isinstance(task, Task) else task.task

    store = ConfigContext.get().store
    with DematerializeRunContext(root_task.flow):
        cached_output, _ = store.retrieve_most_recent_task_output_from_cache(
            root_task, ignore_position_hashes=ignore_position_hashes
        )
        return dematerialize_output_from_store(store, task, cached_output, as_type)


class PseudoStage:
    def __init__(self, stage: "Stage", did_commit: bool):
        self._stage = stage
        self._name = stage.name
        self._transaction_name = stage.name
        self._did_commit = did_commit

        self.id = stage.id

    @property
    def name(self) -> str:
        return self._name

    @property
    def transaction_name(self) -> str:
        return self._transaction_name

    @property
    def current_name(self) -> str:
        if self._did_commit:
            return self.name
        else:
            return self.transaction_name



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.pipedag.context.context import (
    ConfigContext,
    DAGContext,
    StageLockContext,
    TaskContext,
    default_config_dict,
)
from pydiverse.pipedag.context.run_context import (
    FinalTaskState,
    RunContext,
    RunContextServer,
)

__all__ = [
    "DAGContext",
    "TaskContext",
    "ConfigContext",
    "RunContext",
    "RunContextServer",
    "StageLockContext",
    "FinalTaskState",
    "default_config_dict",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import inspect
import random
import re
import time
import types
import typing
import warnings
from dataclasses import dataclass
from typing import Any

import numpy as np
import sqlalchemy as sa
import sqlalchemy.exc
from packaging.version import Version
from pandas.core.dtypes.base import ExtensionDtype

from pydiverse.common import Date, Dtype, PandasBackend
from pydiverse.common.util.computation_tracing import ComputationTracer
from pydiverse.common.util.hashing import stable_hash
from pydiverse.pipedag import ConfigContext
from pydiverse.pipedag._typing import T
from pydiverse.pipedag.backend.table.sql.ddl import (
    ChangeTableLogged,
    CreateTableAsSelect,
    DropTable,
    InsertIntoSelect,
)
from pydiverse.pipedag.backend.table.sql.sql import (
    SQLTableStore,
)
from pydiverse.pipedag.container import ExternalTableReference, Schema, Table
from pydiverse.pipedag.context import TaskContext
from pydiverse.pipedag.errors import HookCheckException
from pydiverse.pipedag.materialize.details import resolve_materialization_details_label
from pydiverse.pipedag.materialize.table_hook_base import (
    AutoVersionSupport,
    CanMatResult,
    CanRetResult,
    TableHook,
)
from pydiverse.pipedag.util.sql import compile_sql

# optional imports
try:
    import polars as pl
except ImportError:
    pl = None

try:
    import dataframely as dy
except ImportError:
    dy = None

try:
    import pydiverse.colspec as cs
except ImportError:
    cs = None

# region SQLALCHEMY
try:
    from sqlalchemy import Select, TextClause
    from sqlalchemy import Text as SqlText
except ImportError:
    # For compatibility with sqlalchemy < 2.0
    from sqlalchemy.sql.expression import TextClause
    from sqlalchemy.sql.selectable import Select

    SqlText = TextClause  # this is what sa.text() returns


def _polars_apply_retrieve_annotation(df, table, store, intentionally_empty: bool = False):
    if cs is not None:
        if inspect.isclass(table.annotation) and issubclass(table.annotation, cs.ColSpec):
            column_spec = table.annotation  # type: cs.ColSpec
            if dy is not None:
                # Try colspec polars specific operation which uses dataframely
                # in the back (casting is not done by SQL based colspec)
                try:
                    df = column_spec.cast_polars(df)
                except pl.exceptions.InvalidOperationError as e:
                    df, failures = column_spec.filter_polars(df, cast=True)
                    with pl.Config() as cfg:
                        cfg.set_tbl_cols(15)
                        cfg.set_tbl_width_chars(120)
                        try:
                            fail_df = str(failures._lf.head(5).collect())
                        except:  # noqa
                            fail_df = str(failures.invalid().head(5))
                    raise HookCheckException(
                        f"Failed casting polars input '{table.name}' to "
                        f"{column_spec.__name__}; "
                        f"Failure counts: {failures.counts()}; "
                        f"\nInvalid:\n{fail_df}"
                    ) from e
    if dy is not None:
        if typing.get_origin(table.annotation) is not None and issubclass(
            typing.get_origin(table.annotation), dy.LazyFrame | dy.DataFrame
        ):
            anno_args = typing.get_args(table.annotation)
            if len(anno_args) == 1:
                column_spec = anno_args[0]
                if issubclass(column_spec, dy.Schema):
                    try:
                        df = column_spec.cast(df)
                    except pl.exceptions.InvalidOperationError as e:
                        df, failures = column_spec.filter(df, cast=True)
                        with pl.Config() as cfg:
                            cfg.set_tbl_cols(15)
                            cfg.set_tbl_width_chars(120)
                            try:
                                fail_df = str(failures._lf.head(5).collect())
                            except:  # noqa
                                fail_df = str(failures.invalid().head(5))
                        raise HookCheckException(
                            f"Failed casting polars input '{table.name}' to "
                            f"{column_spec.__name__}; "
                            f"Failure counts: {failures.counts()}; "
                            f"\nInvalid:\n{fail_df}"
                        ) from e
    return df


def _polars_apply_materialize_annotation(df, table, store):
    if dy is not None:
        column_spec = None
        # support dy.LazyFrame[T] and dy.DataFrame[T] annotations
        if typing.get_origin(table.annotation) is not None and issubclass(
            typing.get_origin(table.annotation), dy.LazyFrame | dy.DataFrame
        ):
            anno_args = typing.get_args(table.annotation)
            if len(anno_args) == 1:
                column_spec = anno_args[0]
        # but also support Schema annotations directly since there is no dy.PdtTable[T]
        if inspect.isclass(table.annotation) and issubclass(table.annotation, dy.Schema):
            column_spec = table.annotation
        if column_spec and issubclass(column_spec, dy.Schema):
            df, failures = column_spec.filter(df, cast=True)
            if len(failures) > 0:
                with pl.Config() as cfg:
                    cfg.set_tbl_cols(15)
                    cfg.set_tbl_width_chars(120)
                    try:
                        fail_df = str(failures._lf.head(5).collect())
                    except:  # noqa
                        fail_df = str(failures.invalid().head(5))
                raise HookCheckException(
                    f"Polars task output {table.name} failed "
                    f"validation with {column_spec.__name__}; "
                    f"Failure counts: {failures.counts()}; "
                    f"\nInvalid:\n{fail_df}"
                )
    if cs is not None:
        if inspect.isclass(table.annotation) and issubclass(table.annotation, cs.ColSpec):
            column_spec = table.annotation
            if dy is not None:
                # Try colspec polars specific operation which uses dataframely
                # in the back
                df, failures = column_spec.filter_polars(df, cast=True)
                if len(failures) > 0:
                    with pl.Config() as cfg:
                        cfg.set_tbl_cols(15)
                        cfg.set_tbl_width_chars(120)
                        try:
                            fail_df = str(failures._lf.head(5).collect())
                        except:  # noqa
                            fail_df = str(failures.invalid().head(5))
                    raise HookCheckException(
                        f"Polars task output {table.name} failed "
                        f"validation with {column_spec.__name__}; "
                        f"Failure counts: {failures.counts()}; "
                        f"\nInvalid:\n{fail_df}"
                    )
            elif pdt_new is not None:
                tbl, failures = column_spec.filter(pdt.Table(df), cast=True)
                if len(failures) > 0:
                    with pl.Config() as cfg:
                        cfg.set_tbl_cols(15)
                        cfg.set_tbl_width_chars(120)
                        debug_invalid_rows = failures.debug_invalid_rows
                        if debug_invalid_rows >> pdt.build_query() and store.engine.dialect.name == "mssql":
                            # this is just a workaround as long as pydiverse.transform
                            # puts an OFFSET in slice_head
                            debug_invalid_rows = failures.debug_invalid_rows >> pdt.arrange(pdt.lit(True))

                        fail_df = str(debug_invalid_rows >> pdt.slice_head(5) >> pdt.export(Polars(lazy=False)))
                    raise HookCheckException(
                        f"Polars task output {table.name} failed "
                        f"validation with {column_spec.__name__}; "
                        f"Failure counts: {failures.counts()}; "
                        f"\nInvalid:\n{fail_df}"
                    )
                df = tbl >> pdt.export(Polars(lazy=False))

    return df


def _sql_apply_materialize_annotation_pdt_early(table: Table, schema: Schema, store: SQLTableStore):
    tbl = table.obj  # type: pdt.Table
    if cs is not None:
        if inspect.isclass(table.annotation) and issubclass(table.annotation, cs.ColSpec):
            column_spec = table.annotation
            if pdt_new is not None:
                # cast columns according to column specification
                # Enum and Struct are currently not supported by pydiverse libraries
                cols = set(c.name for c in tbl)
                tbl = tbl >> pdt.mutate(
                    **{
                        name: tbl[name].cast(col.dtype())
                        for name, col in column_spec.columns().items()
                        if name in cols and not isinstance(col, cs.Enum) and not isinstance(col, cs.Struct)
                    }
                )
    return tbl


def _sql_apply_materialize_annotation(
    table: Table,
    schema: Schema,
    query,
    store: SQLTableStore,
    suffix: str | None,
    unlogged: bool,
) -> tuple[str | None, list[str] | None]:
    invalid_rows = None  # type: str | None
    intermediate_tbls = None  # type: list[str] | None

    def write_pdt_table(tbl: pdt.Table, table_name: str, suffix: str | None = None):
        query = sa.text(str(tbl >> pdt.build_query()))
        schema = store.write_subquery(query, table_name, neighbor_table=table, unlogged=unlogged, suffix=suffix)
        return pdt.Table(table_name, pdt.SqlAlchemy(store.engine, schema=schema.get()))

    def materialize_hook(tbl: pdt.Table, table_prefix):
        name = table_prefix or ""
        name += tbl._ast.name or ""
        name += stable_hash(str(random.randbytes(8)))
        intermediate_tbls.append(name)
        return write_pdt_table(tbl, name)

    if cs is not None:
        if inspect.isclass(table.annotation) and issubclass(table.annotation, cs.ColSpec):
            column_spec = table.annotation
            if pdt_new is not None:
                supported_dialects = ["duckdb", "sqlite", "postgresql", "mssql"]
                if store.engine.dialect.name in supported_dialects:
                    intermediate_tbls = []
                    cfg = cs.config.Config.default
                    cfg.dialect_name = store.engine.dialect.name
                    cfg.materialize_hook = materialize_hook
                    tmp_name = "_raw_" + table.name
                    store.rename_table(table, tmp_name, schema)
                    raw = pdt.Table(tmp_name, pdt.SqlAlchemy(store.engine, schema=schema.get()))
                    tbl, failures = column_spec.filter(raw, cast=True, cfg=cfg)
                    write_pdt_table(tbl, table.name, suffix)
                    failure_counts = failures.counts()
                    if len(failure_counts) > 0:
                        with pl.Config() as cfg:
                            cfg.set_tbl_cols(15)
                            cfg.set_tbl_width_chars(120)
                            if store.engine.dialect.name == "mssql":
                                # this is just a workaround as long as pydiverse.transform
                                # puts an OFFSET in slice_head
                                debug_invalid_rows = failures.debug_invalid_rows >> pdt.arrange(pdt.lit(True))
                            else:
                                debug_invalid_rows = failures.debug_invalid_rows
                            fail_df = str(debug_invalid_rows >> pdt.slice_head(5) >> pdt.export(Polars(lazy=False)))
                        raise HookCheckException(
                            f"Sql task output {table.name} failed "
                            f"validation with {column_spec.__name__}; "
                            f"Failure counts: {failure_counts}; "
                            f"\nInvalid:\n{fail_df}\nQuery:\n{compile_sql(query)}"
                        )
                    invalid_rows = tbl >> pdt.alias(table.name)
                else:
                    store.logger.info(
                        "Colspec annotation ignored because pydiverse.transform currently does not support SQL dialect",
                        dialect=store.engine.dialect.name,
                        supported=supported_dialects,
                    )

    return invalid_rows, intermediate_tbls


@SQLTableStore.register_table()
class SQLAlchemyTableHook(TableHook[SQLTableStore]):
    @dataclass  # consider using pydantic instead
    class Config:
        disable_materialize_annotation_action: bool = False
        disable_retrieve_annotation_action: bool = False
        cleanup_annotation_action_on_success: bool = False
        cleanup_annotation_action_intermediate_state: bool = True
        fault_tolerant_annotation_action: bool = False

    @classmethod
    def cfg(cls) -> Config:
        ret = SQLAlchemyTableHook.Config()
        if hook_args := ConfigContext.get().table_hook_args.get("sql", None):
            for key, value in hook_args.items():
                if hasattr(ret, key):
                    if type(getattr(ret, key)) is not type(value):
                        raise TypeError(
                            f"Invalid type for polars hook argument '{key}': "
                            f"expected {type(getattr(ret, key))}, got {type(value)}"
                        )
                    setattr(ret, key, value)
                else:
                    raise ValueError(f"Unknown polars hook argument '{key}' in table_hook_args.")
        return ret

    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        type_ = type(tbl.obj)
        return CanMatResult.new(issubclass(type_, (sa.sql.expression.TextClause, sa.sql.expression.Selectable)))

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        return CanRetResult.new(type_ == sa.Table)

    @classmethod
    def retrieve_as_reference(cls, type_):
        return True

    @classmethod
    def materialize(
        cls,
        store: SQLTableStore,
        table: Table[sa.sql.expression.TextClause | sa.sql.expression.Selectable],
        stage_name,
        without_config_context: bool = False,
    ):
        query = table.obj
        if isinstance(table.obj, (sa.Table, sa.sql.expression.Alias)):
            query = sa.select("*").select_from(table.obj)
            tbl = table.obj if isinstance(table.obj, sa.Table) else table.obj.original
            source_tables = [
                dict(
                    name=tbl.name,
                    schema=tbl.schema,
                    shared_lock_allowed=table.shared_lock_allowed,
                )
            ]
        else:
            try:
                input_tables = TaskContext.get().input_tables
            except LookupError:
                input_tables = []
            source_tables = [
                dict(
                    name=tbl.name,
                    schema=store.get_schema(tbl.stage.current_name).get()
                    if tbl.external_schema is None
                    else tbl.external_schema,
                    shared_lock_allowed=tbl.shared_lock_allowed,
                )
                for tbl in input_tables
            ]

        schema = store.get_schema(stage_name)

        store.check_materialization_details_supported(resolve_materialization_details_label(table))

        suffix = store.get_create_table_suffix(resolve_materialization_details_label(table))
        unlogged = store.get_unlogged(resolve_materialization_details_label(table))
        if store.dialect_requests_empty_creation(table, is_sql=True):
            cls._create_table_as_select_empty_insert(table, schema, query, source_tables, store, suffix, unlogged)
        else:
            cls._create_table_as_select(table, schema, query, source_tables, store, suffix, unlogged)
        if not without_config_context:
            cfg = cls.cfg()
            invalid_rows, intermediate_tbls = None, None
            if not cfg.disable_materialize_annotation_action:
                try:
                    invalid_rows, intermediate_tbls = _sql_apply_materialize_annotation(
                        table, schema, query, store, suffix, unlogged
                    )
                except Exception as e:  # noqa
                    store.logger.error(
                        "Failed to apply materialize annotation for table",
                        table=table.name,
                        exception=str(e),
                    )
                    if not cfg.fault_tolerant_annotation_action:
                        raise e
            if cfg.cleanup_annotation_action_on_success and invalid_rows is not None:
                store.logger.debug(
                    "Cleaning up intermediate state after successful materialization",
                    table=table.name,
                )
                store.execute(
                    store.execute(DropTable(invalid_rows, schema, if_exists=True)),
                    truncate_printed_select=True,
                )
            if cfg.cleanup_annotation_action_intermediate_state and intermediate_tbls is not None:
                store.logger.debug(
                    "Cleaning up intermediate state after materialization",
                    table=table.name,
                )
                for tbl in intermediate_tbls:
                    store.drop_subquery_table(tbl, schema, neighbor_table=table, if_exists=True)

        store.optional_pause_for_db_transactionality("table_create")

    @classmethod
    def _create_table_as_select(
        cls,
        table: Table,
        schema: Schema,
        query: Select | TextClause | SqlText,
        source_tables: list[dict[str, str]],
        store: SQLTableStore,
        suffix: str,
        unlogged: bool,
    ):
        statements = store.lock_source_tables(source_tables)
        statements += cls._create_as_select_statements(table.name, schema, query, store, suffix, unlogged)
        store.execute(statements)
        store.add_indexes_and_set_nullable(table, schema)

    @classmethod
    def _create_table_as_select_empty_insert(
        cls,
        table: Table,
        schema: Schema,
        query: Select | TextClause | SqlText,
        source_tables: list[dict[str, str]],
        store: SQLTableStore,
        suffix: str,
        unlogged: bool,
    ):
        limit_query = store.get_limit_query(query, rows=0)
        store.execute(cls._create_as_select_statements(table.name, schema, limit_query, store, suffix, unlogged))
        store.add_indexes_and_set_nullable(table, schema, on_empty_table=True)
        statements = store.lock_table(table, schema)
        statements += store.lock_source_tables(source_tables)
        statements += cls._insert_as_select_statements(table.name, schema, query, store)
        store.execute(
            statements,
            truncate_printed_select=True,
        )
        store.add_indexes_and_set_nullable(table, schema, on_empty_table=False)

    @classmethod
    def _create_as_select_statements(
        cls,
        table_name: str,
        schema: Schema,
        query: Select | TextClause | SqlText,
        store: SQLTableStore,
        suffix: str,
        unlogged: bool,
    ):
        _ = store
        return [
            CreateTableAsSelect(
                table_name,
                schema,
                query,
                unlogged=unlogged,
                suffix=suffix,
            )
        ]

    @classmethod
    def _insert_as_select_statements(
        cls,
        table_name: str,
        schema: Schema,
        query: Select | TextClause | SqlText,
        store: SQLTableStore,
    ):
        _ = store
        return [
            InsertIntoSelect(
                table_name,
                schema,
                query,
            )
        ]

    @classmethod
    def retrieve(
        cls,
        store: SQLTableStore,
        table: Table,
        stage_name: str | None,
        as_type: type[sa.Table],
        limit: int | None = None,
    ) -> sa.sql.expression.Selectable:
        assert limit is None, "SQLAlchemyTableHook does not support limit in retrieve."
        table_name, schema = store.resolve_alias(table, stage_name)
        try:
            alias_name = TaskContext.get().name_disambiguator.get_name(table_name)
        except LookupError:
            # Used for imperative materialization with explicit config_context
            alias_name = table_name

        tbl = store.reflect_table(table_name, schema)
        return tbl.alias(alias_name)

    @classmethod
    def lazy_query_str(cls, store, obj) -> str:
        if isinstance(obj, sa.sql.expression.FromClause):
            query = sa.select("*").select_from(obj)
        else:
            query = obj
        query_str = str(query.compile(store.engine, compile_kwargs={"literal_binds": True}))
        # hacky way to canonicalize query (despite __tmp/__even/__odd suffixes
        # and alias resolution)
        query_str = re.sub(r'["\[\]]', "", query_str)
        query_str = re.sub(r'(__tmp|__even|__odd)(?=[ \t\n.;"]|$)', "", query_str.lower())
        return query_str


@SQLTableStore.register_table()
class ExternalTableReferenceHook(TableHook[SQLTableStore]):
    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        type_ = type(tbl.obj)
        return CanMatResult.new(issubclass(type_, ExternalTableReference))

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        return CanRetResult.NO

    @classmethod
    def materialize(
        cls,
        store: SQLTableStore,
        table: Table,
        stage_name: str,
        without_config_context: bool = False,
    ):
        # For an external table reference, we don't need to materialize anything.
        # This is any table referenced by a table reference should already exist
        # in the schema.
        # Instead, we check that the table actually exists.
        stage_schema = store.get_schema(stage_name).get()
        if table.external_schema.upper() == stage_schema.upper():
            raise ValueError(
                f"ExternalTableReference '{table.name}' is not allowed to reference "
                f"tables in the transaction schema '{stage_schema}' of the current "
                "stage."
            )
        if stage_schema.upper().startswith(table.external_schema.upper() + "__"):
            raise ValueError(
                f"ExternalTableReference '{table.name}' is not allowed to reference "
                f"tables in the schema '{table.external_schema}' of the current stage."
            )

        has_table = store.has_table_or_view(table.name, table.external_schema)

        if not has_table:
            raise ValueError(
                f"No table with name '{table.name}' found in schema "
                f"'{table.external_schema}' (reference by ExternalTableReference)."
            )

        return

    @classmethod
    def retrieve(cls, *args, **kwargs):
        raise RuntimeError("This should never get called.")


# endregion

# region PANDAS
try:
    import pandas as pd
except ImportError:
    pd = None


class DataframeSqlTableHook:
    """
    Base class for hooks that handle pandas or polars DataFrames.
    Provides common functionality for uploading and downloading tables.
    """

    @classmethod
    def dialect_has_adbc_driver(cls):
        # by default, we assume an ADBC driver exists
        return True

    @classmethod
    def dialect_supports_connectorx(cls):
        # ConnectorX (used by Polars read_database_uri) does not support many dialects
        return False

    @classmethod
    def download_table(
        cls,
        query: Any,
        store: SQLTableStore,
        dtypes: dict[str, ExtensionDtype | np.dtype] | None = None,
    ) -> Any:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @classmethod
    def upload_table(
        cls,
        table: Table[pd.DataFrame],
        schema: Schema,
        dtypes: dict[str, sa.types.TypeEngine],
        store: SQLTableStore,
        early: bool,
    ):
        """
        Provide hook that allows to override the default
        upload of pandas/polars tables to the tablestore.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @classmethod
    def _get_dialect_dtypes(
        cls, dtypes: dict[str, Dtype], table: Table[pd.DataFrame]
    ) -> dict[str, sa.types.TypeEngine]:
        """
        Convert dtypes to SQLAlchemy types.
        """
        sql_dtypes = {name: dtype.to_sql() for name, dtype in dtypes.items()}
        if table.type_map:
            sql_dtypes.update(table.type_map)
        return sql_dtypes

    @classmethod
    def _dialect_create_empty_table(
        cls,
        store: SQLTableStore,
        table: Table[pd.DataFrame],
        schema: Schema,
        dtypes: dict[str, sa.types.TypeEngine],
    ):
        """
        Create an empty table in the database.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @classmethod
    def get_columns(cls, df):
        return df.columns

    @classmethod
    def _execute_materialize(
        cls,
        table: Table,
        store: SQLTableStore,
        schema: Schema,
        dtypes: dict[str, Dtype],
    ):
        df = table.obj
        dtypes = cls._get_dialect_dtypes(dtypes, table)

        store.check_materialization_details_supported(resolve_materialization_details_label(table))

        if early := store.dialect_requests_empty_creation(table, is_sql=False):
            cls._dialect_create_empty_table(store, table, schema, dtypes)
            store.add_indexes_and_set_nullable(table, schema, on_empty_table=True, table_cols=cls.get_columns(df))
            if store.get_unlogged(resolve_materialization_details_label(table)):
                store.execute(ChangeTableLogged(table.name, schema, False))

        cls.upload_table(table, schema, dtypes, store, early)
        store.add_indexes_and_set_nullable(
            table,
            schema,
            on_empty_table=False if early else None,
            table_cols=cls.get_columns(df),
        )
        store.optional_pause_for_db_transactionality("table_create")

    @classmethod
    def _build_retrieve_query(
        cls,
        store: SQLTableStore,
        table: Table,
        stage_name: str | None,
        limit: int | None = None,
    ) -> tuple[Any, dict[str, Dtype]]:
        table_name, schema = store.resolve_alias(table, stage_name)

        sql_table = store.reflect_table(table_name, schema).alias("tbl")

        cols = {col.name: col for col in sql_table.columns}
        dtypes = {name: Dtype.from_sql(col.type) for name, col in cols.items()}

        cols, dtypes = cls._adjust_cols_retrieve(store, cols, dtypes)

        if cols is None:
            query = sa.select("*").select_from(sql_table)
        else:
            query = sa.select(*cols.values()).select_from(sql_table)
        if limit is not None:
            query = store.get_limit_query(query, rows=limit)
        return query, dtypes

    @classmethod
    def _adjust_cols_retrieve(cls, store: SQLTableStore, cols: dict, dtypes: dict) -> tuple[dict | None, dict]:
        # in earlier times pandas only supported datetime64[ns] and thus we implemented
        # clipping in this function to avoid the creation of dtype=object columns
        return None, dtypes
        # return cols, dtypes


@SQLTableStore.register_table(pd)
class PandasTableHook(TableHook[SQLTableStore], DataframeSqlTableHook):
    """
    Allows overriding the default dtype backend to use by setting the `dtype_backend`
    argument in the `hook_args` section of the table store config::

        hook_args:
          pandas:
            dtype_backend: "arrow" | "numpy"
    """

    pd_version = Version(pd.__version__)
    auto_version_support = AutoVersionSupport.TRACE

    @classmethod
    def upload_table(
        cls,
        table: Table[pd.DataFrame],
        schema: Schema,
        dtypes: dict[str, sa.types.TypeEngine],
        store: SQLTableStore,
        early: bool,
    ):
        """
        Provide hook that allows to override the default
        upload of pandas/polars tables to the tablestore.
        """
        df = table.obj
        schema_name = schema.get()
        with store.engine_connect() as conn:
            with conn.begin():
                if early:
                    store.lock_table(table, schema_name, conn)
                df.to_sql(
                    table.name,
                    conn,
                    schema=schema_name,
                    index=False,
                    dtype=dtypes,
                    chunksize=100_000,
                    if_exists="append" if early else "fail",
                )

    @classmethod
    def download_table(
        cls,
        query: Any,
        store: SQLTableStore,
        dtypes: dict[str, ExtensionDtype | np.dtype] | None = None,
    ) -> pd.DataFrame:
        """
        Provide hook that allows to override the default
        download of pandas tables from the tablestore.
        Also serves as fallback for polars download.
        """
        with store.engine.connect() as conn:
            if PandasTableHook.pd_version >= Version("2.0"):
                df = pd.read_sql(query, con=conn, dtype=dtypes)
            else:
                df = pd.read_sql(query, con=conn)
                df = cls._fix_dtypes(df, dtypes)
            return df

    @classmethod
    def _fix_dtypes(
        cls,
        df: pd.DataFrame,
        dtypes: dict[str, ExtensionDtype | np.dtype] | None = None,
    ) -> pd.DataFrame:
        # df = df.copy(deep=False)  would be costly and is typically not needed
        if dtypes is not None:
            for col, dtype in dtypes.items():
                df[col] = df[col].astype(dtype)
        return df

    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        type_ = type(tbl.obj)
        return CanMatResult.new(issubclass(type_, pd.DataFrame))

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        return CanRetResult.new(type_ == pd.DataFrame)

    @classmethod
    def auto_table(cls, obj: pd.DataFrame):
        if name := obj.attrs.get("name"):
            return Table(obj, name)
        return super().auto_table(obj)

    @classmethod
    def materialize(
        cls,
        store: SQLTableStore,
        table: Table[pd.DataFrame],
        stage_name: str,
        without_config_context: bool = False,
    ):
        df = table.obj.copy(deep=False)
        schema = store.get_schema(stage_name)

        if store.print_materialize:
            store.logger.info(f"Writing table '{schema.get()}.{table.name}'", table_obj=table.obj)

        return cls.materialize_(df, None, store, table, schema)

    @classmethod
    def materialize_(
        cls,
        df: pd.DataFrame,
        dtypes: dict[str:Dtype] | None,
        store: SQLTableStore,
        table: Table[Any],
        schema: Schema,
    ):
        """Helper function that can be invoked by other hooks"""
        if dtypes is None:
            dtypes = {name: Dtype.from_pandas(dtype) for name, dtype in df.dtypes.items()}

        for col, dtype in dtypes.items():
            # Currently, pandas' .to_sql fails for arrow date columns.
            # -> Temporarily convert all dates to objects
            # See: https://github.com/pandas-dev/pandas/issues/53854
            if pd.__version__ < "2.1.3":
                if dtype == Date():
                    df[col] = df[col].astype(object)

        table = table.copy_without_obj()
        table.obj = df

        cls._execute_materialize(
            table=table,
            store=store,
            schema=schema,
            dtypes=dtypes,
        )

    @classmethod
    def _dialect_create_empty_table(
        cls,
        store: SQLTableStore,
        table: Table[pd.DataFrame],
        schema: Schema,
        dtypes: dict[str, sa.types.TypeEngine],
    ):
        df = table.obj  # type: pd.DataFrame
        df[:0].to_sql(
            table.name,
            store.engine,
            schema=schema.get(),
            index=False,
            dtype=dtypes,
        )

    @classmethod
    def retrieve(
        cls,
        store: SQLTableStore,
        table: Table,
        stage_name: str | None,
        as_type: type[pd.DataFrame] | tuple | dict,
        limit: int | None = None,
    ) -> pd.DataFrame:
        # Config
        if PandasTableHook.pd_version >= Version("2.0"):
            # Once arrow is mature enough, we might want to switch to
            # arrow backed dataframes by default
            backend_str = "numpy"
        else:
            backend_str = "numpy"

        try:
            if hook_args := ConfigContext.get().table_hook_args.get("pandas", None):
                if dtype_backend := hook_args.get("dtype_backend", None):
                    backend_str = dtype_backend
        except LookupError:
            pass  # in case dematerialization is called without open ConfigContext

        if isinstance(as_type, tuple):
            backend_str = as_type[1]
        elif isinstance(as_type, dict):
            backend_str = as_type["backend"]

        backend = PandasBackend(backend_str)

        # Retrieve
        query, dtypes = cls._build_retrieve_query(store, table, stage_name, limit)
        dataframe = cls._execute_query_retrieve(store, query, dtypes, backend)
        return dataframe

    @classmethod
    def _execute_query_retrieve(
        cls,
        store: SQLTableStore,
        query: Any,
        dtypes: dict[str, Dtype],
        backend: PandasBackend,
    ) -> pd.DataFrame:
        dtypes = {name: dtype.to_pandas(backend) for name, dtype in dtypes.items()}

        return cls.download_table(query, store, dtypes)

    # Auto Version

    class ComputationTracer(ComputationTracer):
        def _monkey_patch(self):
            import numpy
            import pandas

            from pydiverse.common.util.computation_tracing import patch

            for name in sorted(pandas.__all__):
                try:
                    patch(self, pandas, name)
                except TypeError:
                    pass

            for name in sorted(numpy.__all__):
                try:
                    patch(self, numpy, name)
                except TypeError:
                    pass

    @classmethod
    def get_computation_tracer(cls):
        return cls.ComputationTracer()


# endregion

# region POLARS


@SQLTableStore.register_table(pl)
class PolarsTableHook(TableHook[SQLTableStore], DataframeSqlTableHook):
    @dataclass  # consider using pydantic instead
    class Config:
        disable_materialize_annotation_action: bool = False
        disable_retrieve_annotation_action: bool = False
        fault_tolerant_annotation_action: bool = False

    @classmethod
    def cfg(cls):
        ret = PolarsTableHook.Config()
        if hook_args := ConfigContext.get().table_hook_args.get("polars", None):
            for key, value in hook_args.items():
                if hasattr(ret, key):
                    if type(getattr(ret, key)) is not type(value):
                        raise TypeError(
                            f"Invalid type for polars hook argument '{key}': "
                            f"expected {type(getattr(ret, key))}, got {type(value)}"
                        )
                    setattr(ret, key, value)
                else:
                    raise ValueError(f"Unknown polars hook argument '{key}' in table_hook_args.")
        return ret

    @classmethod
    def _dialect_create_empty_table(
        cls,
        store: SQLTableStore,
        table: Table[pl.DataFrame],
        schema: Schema,
        dtypes: dict[str, sa.types.TypeEngine],
    ):
        df = table.obj  # type: pl.DataFrame
        pd_df = df.to_pandas(use_pyarrow_extension_array=True, zero_copy_only=True)
        pd_df[:0].to_sql(
            table.name,
            store.engine,
            schema=schema.get(),
            index=False,
            dtype=dtypes,
        )
        # _ = dtypes  # unfortunately, we don't know how to set dtypes with polars, yet
        # engine = store.engine
        # table_name = engine.dialect.identifier_preparer.quote(table.name)
        # schema_name = engine.dialect.identifier_preparer.format_schema(schema.get())
        # connection_uri = store.engine_url.render_as_string(hide_password=False)
        # df.slice(0, 0).write_database(
        #     f"{schema_name}.{table_name}",
        #     connection_uri,
        #     if_table_exists="replace",
        #     engine="adbc",
        # )

    @classmethod
    def upload_table(
        cls,
        table: Table[pl.DataFrame],
        schema: Schema,
        dtypes: dict[str, sa.types.TypeEngine],
        store: SQLTableStore,
        early: bool,
    ):
        # use ADBC for writing by default
        df = table.obj  # type: pl.DataFrame
        engine = store.engine
        table_name = engine.dialect.identifier_preparer.quote(table.name)
        schema_name = engine.dialect.identifier_preparer.format_schema(schema.get())
        if not early:
            # as long as we don't know how to provide dtypes to write_database, we need
            # to create the table first
            cls._dialect_create_empty_table(store, table, schema, dtypes)
        if cls.dialect_has_adbc_driver():
            try:
                # try using ADBC, first
                return df.write_database(
                    f"{schema_name}.{table_name}",
                    engine.url.render_as_string(hide_password=False),
                    if_table_exists="append",
                    engine="adbc",
                )
            except Exception as e:  # noqa
                store.logger.warning(
                    "Failed writing table using ADBC, falling back to sqlalchemy: %s",
                    table.name,
                )
        df.write_database(
            f"{schema_name}.{table_name}",
            engine,
            if_table_exists="append",
        )

    @classmethod
    def download_table(
        cls,
        query: Any,
        store: SQLTableStore,
        dtypes: dict[str, pl.DataType] | None = None,
    ) -> pl.DataFrame:
        """
        Provide hook that allows to override the default
        download of polars tables from the tablestore.
        """
        assert dtypes is None, (
            "Polars reads SQL schema and loads the data in reasonable types."
            "Thus, manual dtype manipulation can only done via query or afterwards."
        )
        # We try to use arrow_odbc (see mssql) or adbc (e.g. adbc-driver-postgresql)
        # if possible. Duckdb also has its own implementation.
        connection_uri = store.engine_url.render_as_string(hide_password=False)
        if cls.dialect_has_adbc_driver():
            try:
                df = pl.read_database_uri(query, connection_uri, engine="adbc")
                return cls._fix_dtypes(df, dtypes)
            except:  # noqa
                store.logger.warning(
                    "Failed retrieving query using ADBC, falling back to Pandas: %s",
                    query,
                )
        elif cls.dialect_supports_connectorx():
            # This implementation requires connectorx which does not work for duckdb or ibm_db2.
            # Attention: In case this call fails, we simply fall-back to pandas hook.
            df = pl.read_database_uri(query, connection_uri)
        else:
            # Polars internally falls back to pandas which results in data dependent types
            df = pl.read_database(query, store.engine)
        return cls._fix_dtypes(df, dtypes)

    @classmethod
    def _execute_materialize_polars(
        cls,
        table: Table,
        store: SQLTableStore,
        stage_name: str,
    ):
        df = table.obj
        schema = store.get_schema(stage_name)
        dtypes = {name: Dtype.from_polars(dtype) for name, dtype in df.collect_schema().items()}

        if isinstance(df, pl.LazyFrame):
            df = df.collect()

        table = table.copy_without_obj()
        table.obj = df
        cls._execute_materialize(table, store, schema, dtypes)

    @classmethod
    def _fix_dtypes(cls, df: pl.DataFrame, dtypes: dict[str, pl.DataType] | None = None) -> pl.DataFrame:
        if dtypes is not None:
            return df.with_columns(**{col: df[col].cast(dtype) for col, dtype in dtypes.items()})
        return df

    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        # there is a separate hook for LazyFrame
        type_ = type(tbl.obj)
        # attention: tidypolars.Tibble is subclass of polars DataFrame
        return CanMatResult.new(issubclass(type_, pl.DataFrame))

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        return CanRetResult.new(type_ == pl.DataFrame)

    @classmethod
    def materialize(
        cls,
        store,
        table: Table[pl.DataFrame],
        stage_name: str,
        without_config_context: bool = False,
    ):
        # Materialization for polars happens by first converting the dataframe to
        # a pyarrow backed pandas dataframe, and then calling the PandasTableHook
        # for materialization.

        df = table.obj

        if store.print_materialize:
            schema = store.get_schema(stage_name)
            store.logger.info(f"Writing table '{schema.get()}.{table.name}'", table_obj=table.obj)

        cfg = cls.cfg()
        try:
            if not cfg.disable_materialize_annotation_action:
                df = _polars_apply_materialize_annotation(df, table, store)
            ex = None
        except Exception as e:
            store.logger.error(
                "Failed to apply materialize annotation for table",
                table=table.name,
                exception=str(e),
            )
            ex = e
        finally:
            table = table.copy_without_obj()
            table.obj = df
            cls._execute_materialize_polars(table, store, stage_name)
            if ex and not cfg.fault_tolerant_annotation_action:
                raise ex

    @classmethod
    def retrieve(
        cls,
        store: SQLTableStore,
        table: Table,
        stage_name: str | None,
        as_type: type[pl.DataFrame],
        limit: int | None = None,
    ) -> pl.DataFrame:
        cfg = cls.cfg()
        df = cls._execute_query(store, table, stage_name, as_type, limit=limit)
        if not cfg.disable_retrieve_annotation_action:
            try:
                df = _polars_apply_retrieve_annotation(df, table, store)
            except Exception as e:
                store.logger.error(
                    "Failed to apply retrieve annotation for table",
                    table=table.name,
                    exception=e,
                )
                if not cfg.fault_tolerant_annotation_action:
                    raise e

        df = df.with_columns(pl.col(pl.Datetime).dt.replace_time_zone(None))
        return df

    @classmethod
    def auto_table(cls, obj: pl.DataFrame):
        # currently, we don't know how to store a table name inside polars dataframe
        return super().auto_table(obj)

    @classmethod
    def _compile_query(cls, store: SQLTableStore, query: Select) -> str:
        return str(query.compile(store.engine, compile_kwargs={"literal_binds": True}))

    @classmethod
    def _execute_query(
        cls,
        store: SQLTableStore,
        table: Table,
        stage_name: str,
        as_type: type,
        limit: int | None = None,
    ) -> pl.DataFrame:
        _ = as_type
        query, dtypes = cls._build_retrieve_query(store, table, stage_name, limit)
        # Polars database read methods tend to do a good job in automatic type mapping.
        # Thus dtypes must be corrected after loading if needed.
        dtypes = None
        query = cls._compile_query(store, query)
        fallback_pandas = True
        if cls.dialect_supports_polars_native_read():
            try:
                df = cls.download_table(query, store, dtypes)
                fallback_pandas = False
            except (RuntimeError, ModuleNotFoundError) as e:
                store.logger.error(
                    "Fallback via Pandas since Polars failed to execute query on database %s: %s",
                    store.engine_url.render_as_string(hide_password=True),
                    e,
                )
        if fallback_pandas:
            pandas_hook = store.get_r_table_hook(pd.DataFrame)  # type: PandasTableHook
            backend = PandasBackend.ARROW
            query, dtypes = pandas_hook._build_retrieve_query(store, table, stage_name)
            dtypes = {name: dtype.to_pandas(backend) for name, dtype in dtypes.items()}
            pd_df = pandas_hook.download_table(query, store, dtypes)
            df = pl.from_pandas(pd_df)
        return df

    @classmethod
    def dialect_supports_polars_native_read(cls):
        # for most dialects we find a way
        return True


@SQLTableStore.register_table(pl)
class LazyPolarsTableHook(TableHook[SQLTableStore]):
    auto_version_support = AutoVersionSupport.LAZY

    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        type_ = type(tbl.obj)
        return CanMatResult.new(type_ == pl.LazyFrame)

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        if dy is not None and issubclass(type_, dy.LazyFrame):
            # optionally support input_type=dy.LazyFrame
            return CanRetResult.YES
        return CanRetResult.new(type_ == pl.LazyFrame)

    @classmethod
    def materialize(
        cls,
        store,
        table: Table[pl.LazyFrame],
        stage_name,
        without_config_context: bool = False,
    ):
        t = table.obj
        table = table.copy_without_obj()
        table.obj = t.collect()

        polars_hook = store.get_m_table_hook(table)
        return polars_hook.materialize(store, table, stage_name)

    @classmethod
    def retrieve(
        cls,
        store: SQLTableStore,
        table: Table,
        stage_name: str | None,
        as_type: type[pl.DataFrame],
        limit: int | None = None,
    ) -> pl.LazyFrame:
        polars_hook = store.get_r_table_hook(pl.DataFrame)
        result = polars_hook.retrieve(store, table, stage_name, as_type, limit)

        return result.lazy()

    @classmethod
    def retrieve_for_auto_versioning_lazy(
        cls,
        store: SQLTableStore,
        table: Table,
        stage_name: str,
        as_type: type[pl.LazyFrame],
    ) -> pl.LazyFrame:
        polars_hook = store.get_r_table_hook(pl.DataFrame)  # type: PolarsTableHook
        df = polars_hook.retrieve(store, table, stage_name, as_type, limit=0)

        # Create lazy frame where each column is identified by:
        #     stage name, table name, column name
        # We then rename all columns to match the names of the table.
        #
        # This allows us to properly trace the origin of each column in
        # the output `.serialize` back to the table where it originally came from.

        schema = {}
        rename = {}
        for col in df:
            qualified_name = f"[{table.stage.name}].[{table.name}].[{col.name}]"
            schema[qualified_name] = col.dtype
            rename[qualified_name] = col.name

        lf = pl.LazyFrame(schema=schema).rename(rename)
        return lf

    @classmethod
    def get_auto_version_lazy(cls, obj) -> str:
        """
        :param obj: object returned from task
        :return: string representation of the operations performed on this object.
        :raises TypeError: if the object doesn't support automatic versioning.
        """
        if not isinstance(obj, pl.LazyFrame):
            raise TypeError("Expected LazyFrame")
        return str(obj.serialize())


try:
    import tidypolars
    from tidypolars import Tibble
except ImportError as e:
    warnings.warn(str(e), ImportWarning)
    tidypolars = None
    Tibble = None


@SQLTableStore.register_table(tidypolars, pl)
class TidyPolarsTableHook(TableHook[SQLTableStore]):
    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        type_ = type(tbl.obj)
        return CanMatResult.new(issubclass(type_, Tibble))

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        return CanRetResult.new(type_ == Tibble)

    @classmethod
    def materialize(
        cls,
        store,
        table: Table[Tibble],
        stage_name,
        without_config_context: bool = False,
    ):
        warnings.warn(
            "tidypolars support is deprecated since tidypolars does not work with current version of polars",
            DeprecationWarning,
            stacklevel=2,
        )
        t = table.obj
        table = table.copy_without_obj()
        table.obj = t.to_polars()

        polars_hook = store.get_m_table_hook(table)
        return polars_hook.materialize(store, table, stage_name)

    @classmethod
    def retrieve(
        cls,
        store: SQLTableStore,
        table: Table,
        stage_name: str | None,
        as_type: type[Tibble],
        limit: int | None = None,
    ) -> Tibble:
        warnings.warn(
            "tidypolars support is deprecated since tidypolars does not work with current version of polars",
            DeprecationWarning,
            stacklevel=2,
        )
        polars_hook = store.get_r_table_hook(pl.DataFrame)
        df = polars_hook.retrieve(store, table, stage_name, as_type, limit)
        return tidypolars.from_polars(df)

    @classmethod
    def auto_table(cls, obj: Tibble):
        # currently, we don't know how to store a table name inside tidypolar   s tibble
        return super().auto_table(obj)


# endregion

# region PYDIVERSE TRANSFORM

try:
    # optional dependency to pydiverse-transform
    import pydiverse.transform as pdt

    try:
        # Detect which version of pdtransform is installed
        # this import only exists in version <0.2
        from pydiverse.transform.eager import PandasTableImpl

        # ensures a "used" state for the import, preventing black from deleting it
        _ = PandasTableImpl

        pdt_old = pdt
        pdt_new = None
    except ImportError:
        try:
            # detect if 0.2 or >0.2 is active
            # this import would only work in <=0.2
            from pydiverse.transform.extended import Polars

            # ensures a "used" state for the import, preventing black from deleting it
            _ = Polars

            pdt_old = None
            pdt_new = pdt
        except ImportError:
            raise NotImplementedError("pydiverse.transform 0.2.0 - 0.2.2 isn't supported") from None
except ImportError as e:
    warnings.warn(str(e), ImportWarning)
    pdt = types.ModuleType("pydiverse.transform")
    pdt.Table = None
    pdt_old = None
    pdt_new = None


@SQLTableStore.register_table(pdt_old)
class PydiverseTransformTableHookOld(TableHook[SQLTableStore]):
    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        type_ = type(tbl.obj)
        return CanMatResult.new(issubclass(type_, pdt.Table))

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        from pydiverse.transform.eager import PandasTableImpl
        from pydiverse.transform.lazy import SQLTableImpl

        return CanRetResult.new(issubclass(type_, (PandasTableImpl, SQLTableImpl)))

    @classmethod
    def retrieve_as_reference(cls, type_) -> bool:
        from pydiverse.transform.lazy import SQLTableImpl

        return issubclass(type_, SQLTableImpl)

    @classmethod
    def materialize(
        cls,
        store,
        table: Table[pdt.Table],
        stage_name,
        without_config_context: bool = False,
    ):
        from pydiverse.transform.core.verbs import collect
        from pydiverse.transform.eager import PandasTableImpl
        from pydiverse.transform.lazy import SQLTableImpl

        t = table.obj
        table = table.copy_without_obj()
        if isinstance(t._impl, PandasTableImpl):
            table.obj = t >> collect()
            hook = store.get_m_table_hook(table)
            return hook.materialize(store, table, stage_name)
        if isinstance(t._impl, SQLTableImpl):
            table.obj = t._impl.build_select()
            hook = store.get_m_table_hook(table)
            return hook.materialize(store, table, stage_name)
        raise NotImplementedError

    @classmethod
    def retrieve(
        cls,
        store: SQLTableStore,
        table: Table,
        stage_name: str | None,
        as_type: type[T],
        limit: int | None = None,
    ) -> T:
        from pydiverse.transform.eager import PandasTableImpl
        from pydiverse.transform.lazy import SQLTableImpl

        if issubclass(as_type, PandasTableImpl):
            hook = store.get_r_table_hook(pd.DataFrame)
            df = hook.retrieve(store, table, stage_name, pd.DataFrame, limit)
            return pdt.Table(PandasTableImpl(table.name, df))
        if issubclass(as_type, SQLTableImpl):
            hook = store.get_r_table_hook(sa.Table)
            sa_tbl = hook.retrieve(store, table, stage_name, sa.Table, limit)
            return pdt.Table(SQLTableImpl(store.engine, sa_tbl))
        raise NotImplementedError

    @classmethod
    def auto_table(cls, obj: pdt.Table):
        return Table(obj, obj._impl.name)

    @classmethod
    def lazy_query_str(cls, store, obj: pdt.Table) -> str:
        from pydiverse.transform.core.verbs import build_query

        query = obj >> build_query()

        if query is not None:
            return str(query)
        return super().lazy_query_str(store, obj)


@SQLTableStore.register_table(pdt_new)
class PydiverseTransformTableHook(TableHook[SQLTableStore]):
    auto_version_support = AutoVersionSupport.LAZY

    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        type_ = type(tbl.obj)
        return CanMatResult.new(issubclass(type_, pdt.Table))

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        from pydiverse.transform.extended import (
            Pandas,
            Polars,
            SqlAlchemy,
        )

        return CanRetResult.new(issubclass(type_, (Polars, SqlAlchemy, Pandas)))

    @classmethod
    def retrieve_as_reference(cls, type_) -> bool:
        from pydiverse.transform.extended import SqlAlchemy

        return issubclass(type_, SqlAlchemy)

    @classmethod
    def materialize(
        cls,
        store,
        table: Table[pdt.Table],
        stage_name,
        without_config_context: bool = False,
    ):
        from pydiverse.transform.extended import (
            Polars,
            build_query,
            export,
        )

        t = table.obj
        table_cpy = table.copy_without_obj()
        query = t >> build_query()
        # detect SQL by checking whether build_query() succeeds
        if query is not None:
            # continue with SQL case handling
            table_cpy.obj = sa.text(str(query))
            hook = store.get_m_table_hook(table_cpy)
            assert hook, "fatal error: no hook for materialization of SqlAlchemy query found"
            cfg = hook.cfg()
            if not cfg.disable_materialize_annotation_action:
                schema = store.get_schema(stage_name)
                t = _sql_apply_materialize_annotation_pdt_early(table, schema, store)
                table_cpy.obj = sa.text(str(t >> build_query()))
            return hook.materialize(store, table_cpy, stage_name)
        else:
            # use Polars for dataframe handling
            table_cpy.obj = t >> export(Polars(lazy=True))
            hook = store.get_m_table_hook(table_cpy)
            return hook.materialize(store, table_cpy, stage_name)

    @classmethod
    def retrieve(
        cls,
        store: SQLTableStore,
        table: Table,
        stage_name: str | None,
        as_type: type[T],
        limit: int | None = None,
    ) -> T:
        from pydiverse.transform.extended import Pandas, Polars, SqlAlchemy

        if issubclass(as_type, Polars):
            import polars as pl

            hook = store.get_r_table_hook(pl.LazyFrame)
            lf = hook.retrieve(store, table, stage_name, pl.DataFrame, limit)
            return pdt.Table(lf, name=table.name)
        elif issubclass(as_type, SqlAlchemy):
            hook = store.get_r_table_hook(sa.Table)
            sa_tbl = hook.retrieve(store, table, stage_name, sa.Table, limit)
            return pdt.Table(
                sa_tbl.original.name,
                SqlAlchemy(store.engine, schema=sa_tbl.original.schema),
            )
        elif issubclass(as_type, Pandas):
            import pandas as pd

            hook = store.get_r_table_hook(pd.DataFrame)
            df = hook.retrieve(store, table, stage_name, pd.DataFrame, limit)
            return pdt.Table(df, name=table.name)

        raise NotImplementedError

    @classmethod
    def auto_table(cls, obj: pdt.Table):
        return Table(obj, obj._ast.name)

    @classmethod
    def lazy_query_str(cls, store, obj: pdt.Table) -> str:
        from pydiverse.transform.extended import (
            build_query,
        )

        query = obj >> build_query()

        if query is not None:
            return str(query)
        return super().lazy_query_str(store, obj)

    @classmethod
    def retrieve_for_auto_versioning_lazy(
        cls,
        store: SQLTableStore,
        table: Table,
        stage_name: str,
        as_type: type[pdt.Table],
    ) -> pdt.Table:
        polars_hook = store.get_r_table_hook(pl.LazyFrame)  # type: PolarsTableHook
        lf = polars_hook.retrieve_for_auto_versioning_lazy(store, table, stage_name, as_type)
        return pdt.Table(lf, name=table.name)

    @classmethod
    def get_auto_version_lazy(cls, obj) -> str:
        from pydiverse.transform.extended import (
            build_query,
        )

        query = obj >> build_query()

        if query is not None:
            # This is intended to throw a TypeError because version=AUTO_VERSION
            # does not make sense for SQL tasks; use lazy=True instead
            return super().get_auto_version_lazy(obj)

        return LazyPolarsTableHook.get_auto_version_lazy(obj >> pdt.export(pdt.Polars(lazy=True)))


# endregion

# region IBIS

try:
    import ibis
except ImportError as e:
    warnings.warn(str(e), ImportWarning)
    ibis = types.ModuleType("ibis")
    ibis.api = types.ModuleType("ibis.api")
    ibis.api.Table = None


@SQLTableStore.register_table(ibis.api.Table)
class IbisTableHook(TableHook[SQLTableStore]):
    @classmethod
    def conn(cls, store: SQLTableStore):
        if conn := store.hook_cache.get((cls, "conn")):
            return conn
        conn = cls._conn(store)
        store.hook_cache[(cls, "conn")] = conn
        return conn

    @classmethod
    def _conn(cls, store: SQLTableStore):
        return ibis.connect(store.engine_url.render_as_string(hide_password=False))

    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        type_ = type(tbl.obj)
        # Operations on a table like mutate() or join() don't change the type
        return CanMatResult.new(issubclass(type_, ibis.api.Table))

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        return CanRetResult.new(issubclass(type_, ibis.api.Table))

    @classmethod
    def retrieve_as_reference(cls, type_) -> bool:
        return True

    @classmethod
    def materialize(
        cls,
        store,
        table: Table[ibis.api.Table],
        stage_name,
        without_config_context: bool = False,
    ):
        t = table.obj
        table = table.copy_without_obj()
        table.obj = sa.text(cls.lazy_query_str(store, t))

        sa_hook = store.get_m_table_hook(table)
        return sa_hook.materialize(store, table, stage_name)

    @classmethod
    def retrieve(
        cls,
        store: SQLTableStore,
        table: Table,
        stage_name: str | None,
        as_type: type[ibis.api.Table],
        limit: int | None = None,
    ) -> ibis.api.Table:
        assert limit is None, "IbisTableHook does not support limit in retrieve."
        conn = cls.conn(store)
        table_name, schema = store.resolve_alias(table, stage_name)
        for retry_iteration in range(4):
            # retry operation since it might have been terminated as a deadlock victim
            try:
                tbl = conn.table(
                    table_name,
                    database=schema,
                )
                break
            except (sa.exc.SQLAlchemyError, sa.exc.DBAPIError):
                if retry_iteration == 3:
                    raise
                time.sleep(retry_iteration * retry_iteration * 1.2)
        else:
            raise Exception
        return tbl

    @classmethod
    def auto_table(cls, obj: ibis.api.Table):
        if obj.has_name():
            return Table(obj, obj.get_name())
        else:
            return super().auto_table(obj)

    @classmethod
    def lazy_query_str(cls, store, obj: ibis.api.Table) -> str:
        return str(ibis.to_sql(obj, cls.conn(store).name))


# endregion
